@@ -62,7 +62,6 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 	private int _maxNumThreads = -1; //-1 for unlimited
 	
 	//removeEmpty hints
-	private boolean _outputEmptyBlocks = true;
 	private boolean _outputPermutationMatrix = false;
 	private boolean _bRmEmptyBC = false;
 	
@@ -74,7 +73,7 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 	 * i.e., getInput().get(_paramIndexMap.get(parameterName)) refers to the Hop
 	 * that is associated with parameterName.
 	 */
-	private HashMap<String, Integer> _paramIndexMap = new HashMap<String, Integer>();
+	private HashMap<String, Integer> _paramIndexMap = new HashMap<>();
 
 	private ParameterizedBuiltinOp() {
 		//default constructor for clone
@@ -128,10 +127,6 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 	public ParamBuiltinOp getOp() {
 		return _op;
 	}
-
-	public void setOutputEmptyBlocks(boolean flag) {
-		_outputEmptyBlocks = flag;
-	}
 	
 	public void setOutputPermutationMatrix(boolean flag) {
 		_outputPermutationMatrix = flag;
@@ -170,7 +165,7 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 			return getLops();
 		
 		// construct lops for all input parameters
-		HashMap<String, Lop> inputlops = new HashMap<String, Lop>();
+		HashMap<String, Lop> inputlops = new HashMap<>();
 		for (Entry<String, Integer> cur : _paramIndexMap.entrySet())
 			inputlops.put(cur.getKey(), getInput().get(cur.getValue()).constructLops());
 
@@ -194,9 +189,10 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 			case CDF:
 			case INVCDF: 
 			case REPLACE:
-			case TRANSFORMAPPLY: 
-			case TRANSFORMDECODE: 
-			case TRANSFORMMETA: 
+			case TRANSFORMAPPLY:
+			case TRANSFORMDECODE:
+			case TRANSFORMCOLMAP:
+			case TRANSFORMMETA:
 			case TOSTRING: {
 				ExecType et = optFindExecType();
 				ParameterizedBuiltin pbilop = new ParameterizedBuiltin(inputlops,
@@ -322,7 +318,7 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 						long m2_dim1 = groups.getDim1();
 						long m2_dim2 = groups.getDim2();
 						long m3_dim1 = m1_dim1; 
-						long m3_dim2 = ((m1_dim2>0 && m2_dim2>0) ? (m1_dim2 + m2_dim2) : -1);
+						long m3_dim2 = ((m1_dim2>=0 && m2_dim2>=0) ? (m1_dim2 + m2_dim2) : -1);
 						long m3_nnz = (target.getNnz()>0 && groups.getNnz()>0) ? (target.getNnz() + groups.getNnz()) : -1; 
 						long brlen = target.getRowsInBlock();
 						long bclen = target.getColsInBlock();
@@ -412,6 +408,7 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 		Hop targetHop = getTargetHop();
 		Hop marginHop = getParameterHop("margin");
 		Hop selectHop = getParameterHop("select");
+		Hop emptyRet = getParameterHop("empty.return");
 		
 		if( et == ExecType.CP || et == ExecType.CP_FILE )
 		{
@@ -478,8 +475,8 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 		else if( et == ExecType.MR )
 		{
 			//special compile for mr removeEmpty-diag 
-			if(    isTargetDiagInput() && marginHop instanceof LiteralOp 
-				&& ((LiteralOp)marginHop).getStringValue().equals("rows") )
+			if( isTargetDiagInput()
+				&& HopRewriteUtils.isLiteralOfValue(marginHop, "rows") )
 	 		{
 				//get input vector (without materializing diag())
 				Hop input = targetHop.getInput().get(0);
@@ -606,13 +603,14 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 				Lop rmEmpty = null;
 				
 				//a) broadcast-based PMM (permutation matrix mult)
-				if( rmRows && rlen > 0 && mestPM < OptimizerUtils.getRemoteMemBudgetMap() )
+				if( rmRows && rlen >= 0 && mestPM < OptimizerUtils.getRemoteMemBudgetMap()
+					&& HopRewriteUtils.isLiteralOfValue(emptyRet, false))
 				{
 					boolean needPart = !offsets.dimsKnown() || offsets.getDim1() > DistributedCacheInput.PARTITION_SIZE;
 					if( needPart ){ //requires partitioning
 						loffset = new DataPartition(loffset, DataType.MATRIX, ValueType.DOUBLE, (mestPM>OptimizerUtils.getLocalMemBudget())?ExecType.MR:ExecType.CP, PDataPartitionFormat.ROW_BLOCK_WISE_N);
 						loffset.getOutputParameters().setDimensions(rlen, 1, brlen, bclen, rlen);
-						setLineNumbers(loffset);	
+						setLineNumbers(loffset);
 					}
 					
 					rmEmpty = new PMMJ(loffset, linput, lmaxdim, getDataType(), getValueType(), needPart, true, ExecType.MR);
@@ -622,8 +620,8 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 				//b) general case: repartition-based rmempty
 				else
 				{
-					boolean requiresRep =   ((clen>bclen || clen<=0) &&  rmRows) 
-							             || ((rlen>brlen || rlen<=0) && !rmRows);
+					boolean requiresRep = ((clen>bclen || clen<=0) &&  rmRows)
+						|| ((rlen>brlen || rlen<=0) && !rmRows);
 					
 					if( requiresRep ) {
 						Lop pos = createOffsetLop(input, rmRows); //ncol of left input (determines num replicates)
@@ -640,11 +638,12 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 					setLineNumbers(group2);
 					group2.getOutputParameters().setDimensions(rlen, clen, brlen, bclen, nnz);
 				
-					HashMap<String, Lop> inMap = new HashMap<String, Lop>();
+					HashMap<String, Lop> inMap = new HashMap<>();
 					inMap.put("target", group1);
 					inMap.put("offset", group2);
 					inMap.put("maxdim", lmaxdim);
 					inMap.put("margin", inputlops.get("margin"));
+					inMap.put("empty.return", inputlops.get("empty.return"));
 					
 					rmEmpty = new ParameterizedBuiltin(inMap, HopsParameterizedBuiltinLops.get(_op), getDataType(), getValueType(), et);			
 					setOutputDimensions(rmEmpty);
@@ -725,12 +724,13 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 			Lop loffset = offsets.constructLops();
 			Lop lmaxdim = maxDim.constructLops();
 			
-			HashMap<String, Lop> inMap = new HashMap<String, Lop>();
+			HashMap<String, Lop> inMap = new HashMap<>();
 			inMap.put("target", linput);
 			inMap.put("offset", loffset);
 			inMap.put("maxdim", lmaxdim);
 			inMap.put("margin", inputlops.get("margin"));
-		
+			inMap.put("empty.return", inputlops.get("empty.return"));
+			
 			if ( !FORCE_DIST_RM_EMPTY && isRemoveEmptyBcSP())
 				_bRmEmptyBC = true;
 			
@@ -1070,6 +1070,7 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 		//force CP for in-memory only transform builtins
 		if( (_op == ParamBuiltinOp.TRANSFORMAPPLY && REMOTE==ExecType.MR)
 			|| _op == ParamBuiltinOp.TRANSFORMDECODE && REMOTE==ExecType.MR
+			|| _op == ParamBuiltinOp.TRANSFORMCOLMAP 
 			|| _op == ParamBuiltinOp.TRANSFORMMETA 
 			|| _op == ParamBuiltinOp.TOSTRING 
 			|| _op == ParamBuiltinOp.CDF 
@@ -1162,13 +1163,18 @@ public class ParameterizedBuiltinOp extends Hop implements MultiThreadedHop
 				//TODO parse json spec
 				break;
 			}
-			
 			case TRANSFORMAPPLY: {
 				//rows remain unchanged only if no omitting
 				//cols remain unchanged of no dummy coding 
 				//TODO parse json spec
 				break;
 			}
+			case TRANSFORMCOLMAP: {
+				Hop target = getTargetHop();
+				setDim1( target.getDim2() );
+				setDim2( 3 ); //fixed schema
+				break;
+			}
 			default:
 				//do nothing
 				break;
