@@ -60,7 +60,6 @@ import org.apache.sysml.runtime.instructions.cp.UnaryCPInstruction;
 import org.apache.sysml.runtime.instructions.cp.VariableCPInstruction;
 import org.apache.sysml.runtime.instructions.mr.MRInstruction;
 import org.apache.sysml.runtime.matrix.MatrixCharacteristics;
-import org.apache.sysml.runtime.matrix.MatrixDimensionsMetaData;
 import org.apache.sysml.runtime.matrix.operators.CMOperator;
 import org.apache.sysml.runtime.matrix.operators.CMOperator.AggregateOperationTypes;
 import org.apache.sysml.runtime.util.UtilFunctions;
@@ -217,7 +216,7 @@ public abstract class CostEstimator
 		return ret;
 	}
 	
-	private void maintainVariableStatistics( LocalVariableMap vars, HashMap<String, VarStats> stats ) 
+	private static void maintainVariableStatistics( LocalVariableMap vars, HashMap<String, VarStats> stats ) 
 		throws DMLRuntimeException
 	{
 		for( String varname : vars.keySet() )
@@ -227,7 +226,7 @@ public abstract class CostEstimator
 			if( dat instanceof MatrixObject ) //matrix
 			{
 				MatrixObject mo = (MatrixObject) dat;
-				MatrixCharacteristics mc = ((MatrixDimensionsMetaData)mo.getMetaData()).getMatrixCharacteristics();
+				MatrixCharacteristics mc = mo.getMatrixCharacteristics();
 				long rlen = mc.getRows();
 				long clen = mc.getCols();
 				long brlen = mc.getRowsPerBlock();
@@ -242,11 +241,10 @@ public abstract class CostEstimator
 			}
 			
 			stats.put(varname, vs);
-			//System.out.println(varname+" "+vs);
 		}
 	}
 	
-	private void maintainCPInstVariableStatistics( CPInstruction inst, HashMap<String, VarStats> stats )
+	private static void maintainCPInstVariableStatistics( CPInstruction inst, HashMap<String, VarStats> stats )
 	{
 		if( inst instanceof VariableCPInstruction )
 		{
@@ -264,8 +262,6 @@ public abstract class CostEstimator
 				long nnz = Long.parseLong(parts[10]);
 				VarStats vs = new VarStats(rlen, clen, brlen, bclen, nnz, false);
 				stats.put(varname, vs);
-				
-				//System.out.println(varname+" "+vs);
 			}
 			else if ( optype.equals("cpvar") ) {
 				String varname = parts[1];
@@ -308,10 +304,7 @@ public abstract class CostEstimator
 			FunctionCallCPInstruction finst = (FunctionCallCPInstruction) inst;
 			ArrayList<String> outVars = finst.getBoundOutputParamNames();
 			for( String varname : outVars )
-			{
 				stats.put(varname, _unknownStats);
-				//System.out.println(varname+" "+vs);
-			}
 		}
 	}
 
@@ -351,7 +344,7 @@ public abstract class CostEstimator
 		}
 		
 		//compute intermediate result indices
-		HashMap<Byte,MatrixCharacteristics> dims = new HashMap<Byte, MatrixCharacteristics>();
+		HashMap<Byte,MatrixCharacteristics> dims = new HashMap<>();
 		//populate input indices
 		for( Entry<String,VarStats> e : stats.entrySet() )
 		{
@@ -427,7 +420,7 @@ public abstract class CostEstimator
 		return ret;
 	}
 	
-	private Object[] extractCPInstStatistics( Instruction inst, HashMap<String, VarStats> stats )
+	private static Object[] extractCPInstStatistics( Instruction inst, HashMap<String, VarStats> stats )
 	{		
 		Object[] ret = new Object[2]; //stats, attrs
 		VarStats[] vs = new VarStats[3];
@@ -575,13 +568,13 @@ public abstract class CostEstimator
 		return ret;
 	}
 	
-	private void setUnknownStats(VarStats[] vs) {
+	private static void setUnknownStats(VarStats[] vs) {
 		vs[0] = _unknownStats;
 		vs[1] = _unknownStats;
 		vs[2] = _unknownStats;	
 	}
 	
-	private Object[] extractMRJobInstStatistics( Instruction inst, HashMap<String, VarStats> stats )
+	private static Object[] extractMRJobInstStatistics( Instruction inst, HashMap<String, VarStats> stats )
 	{
 		Object[] ret = new Object[2]; //stats, attrs
 		VarStats[] vs = null;
@@ -614,7 +607,7 @@ public abstract class CostEstimator
 		return ret;
 	}	
 	
-	private void cleanupMRJobVariableStatistics( Instruction inst, HashMap<String, VarStats> stats )
+	private static void cleanupMRJobVariableStatistics( Instruction inst, HashMap<String, VarStats> stats )
 	{
 		MRJobInstruction jinst = (MRJobInstruction)inst;
 		
@@ -634,7 +627,7 @@ public abstract class CostEstimator
 		}
 	}
 	
-	private long getNumIterations(HashMap<String,VarStats> stats, ForProgramBlock pb) {
+	private static long getNumIterations(HashMap<String,VarStats> stats, ForProgramBlock pb) {
 		return OptimizerUtils.getNumIterations(pb, DEFAULT_NUMITER);
 	}
 
