@@ -58,7 +58,7 @@ import org.apache.sysml.runtime.instructions.cp.Data;
 import org.apache.sysml.runtime.instructions.cp.ScalarObject;
 import org.apache.sysml.runtime.instructions.cp.ScalarObjectFactory;
 import org.apache.sysml.runtime.matrix.MatrixCharacteristics;
-import org.apache.sysml.runtime.matrix.MatrixFormatMetaData;
+import org.apache.sysml.runtime.matrix.MetaDataFormat;
 import org.apache.sysml.udf.lib.DynamicReadMatrixCP;
 import org.apache.sysml.udf.lib.DynamicReadMatrixRcCP;
 import org.apache.sysml.udf.lib.OrderWrapper;
@@ -84,7 +84,7 @@ public class InterProceduralAnalysis
 {
 	private static final boolean LDEBUG = false; //internal local debug level
 	private static final Log LOG = LogFactory.getLog(InterProceduralAnalysis.class.getName());
-    
+
 	//internal configuration parameters
 	protected static final boolean INTRA_PROCEDURAL_ANALYSIS      = true; //propagate statistics across statement blocks (main/functions)	
 	protected static final boolean PROPAGATE_KNOWN_UDF_STATISTICS = true; //propagate statistics for known external functions 
@@ -96,12 +96,13 @@ public class InterProceduralAnalysis
 	protected static final boolean PROPAGATE_SCALAR_VARS_INTO_FUN = true; //propagate scalar variables into functions that are called once
 	protected static final boolean PROPAGATE_SCALAR_LITERALS      = true; //propagate and replace scalar literals into functions
 	protected static final boolean APPLY_STATIC_REWRITES          = true; //apply static hop dag and statement block rewrites
+	protected static final int     INLINING_MAX_NUM_OPS           = 10;    //inline single-statement functions w/ #ops <= threshold, other than dataops and literals
 	
 	static {
 		// for internal debugging only
 		if( LDEBUG ) {
 			Logger.getLogger("org.apache.sysml.hops.ipa.InterProceduralAnalysis")
-				  .setLevel((Level) Level.DEBUG);
+				.setLevel((Level) Level.DEBUG);
 		}
 	}
 	
@@ -129,13 +130,14 @@ public class InterProceduralAnalysis
 		_fgraph = new FunctionCallGraph(dmlp);
 		
 		//create order list of IPA passes
-		_passes = new ArrayList<IPAPass>();
+		_passes = new ArrayList<>();
 		_passes.add(new IPAPassRemoveUnusedFunctions());
 		_passes.add(new IPAPassFlagFunctionsRecompileOnce());
 		_passes.add(new IPAPassRemoveUnnecessaryCheckpoints());
 		_passes.add(new IPAPassRemoveConstantBinaryOps());
 		_passes.add(new IPAPassPropagateReplaceLiterals());
 		_passes.add(new IPAPassApplyStaticHopRewrites());
+		_passes.add(new IPAPassInlineFunctions());
 	}
 	
 	public InterProceduralAnalysis(StatementBlock sb) {
@@ -145,7 +147,7 @@ public class InterProceduralAnalysis
 		_fgraph = new FunctionCallGraph(sb);
 		
 		//create order list of IPA passes
-		_passes = new ArrayList<IPAPass>();
+		_passes = new ArrayList<>();
 	}
 	
 	/**
@@ -164,13 +166,14 @@ public class InterProceduralAnalysis
 	 * @throws HopsException in case of compilation errors
 	 */
 	public void analyzeProgram(int repetitions) 
-		throws HopsException	
+		throws HopsException
 	{
 		//sanity check for valid number of repetitions
 		if( repetitions <= 0 )
 			throw new HopsException("Invalid number of IPA repetitions: " + repetitions);
 		
 		//perform number of requested IPA iterations
+		FunctionCallSizeInfo lastSizes = null;
 		for( int i=0; i<repetitions; i++ ) {
 			if( LOG.isDebugEnabled() )
 				LOG.debug("IPA: start IPA iteration " + (i+1) + "/" + repetitions +".");
@@ -198,14 +201,23 @@ public class InterProceduralAnalysis
 			
 			//step 2: apply additional IPA passes
 			for( IPAPass pass : _passes )
-				if( pass.isApplicable() )
+				if( pass.isApplicable(_fgraph) )
 					pass.rewriteProgram(_prog, _fgraph, fcallSizes);
+			
+			//early abort without functions or on reached fixpoint
+			if( _fgraph.getReachableFunctions().isEmpty() 
+				|| (lastSizes != null && lastSizes.equals(fcallSizes)) ) {
+				if( LOG.isDebugEnabled() )
+					LOG.debug("IPA: Early abort after " + (i+1) + "/" + repetitions
+						+ " repetitions due to reached fixpoint.");
+				break;
+			}
 		}
 		
 		//cleanup pass: remove unused functions
 		FunctionCallGraph graph2 = new FunctionCallGraph(_prog);
 		IPAPass rmFuns = new IPAPassRemoveUnusedFunctions();
-		if( rmFuns.isApplicable() )
+		if( rmFuns.isApplicable(graph2) )
 			rmFuns.rewriteProgram(_prog, graph2, null);
 	}
 	
@@ -241,7 +253,7 @@ public class InterProceduralAnalysis
 		//check size-preserving characteristic
 		if( ret ) {
 			FunctionCallSizeInfo fcallSizes = new FunctionCallSizeInfo(_fgraph, false);
-			HashSet<String> fnStack = new HashSet<String>();
+			HashSet<String> fnStack = new HashSet<>();
 			LocalVariableMap callVars = new LocalVariableMap();
 			
 			//populate input
@@ -341,7 +353,7 @@ public class InterProceduralAnalysis
 			//remove updated constant scalars
 			Recompiler.removeUpdatedScalars(callVars, sb);
 			//old stats in, new stats out if updated
-			ArrayList<Hop> roots = sb.get_hops();
+			ArrayList<Hop> roots = sb.getHops();
 			DMLProgram prog = sb.getDMLProg();
 			//replace scalar reads with literals
 			Hop.resetVisitStatus(roots);
@@ -370,7 +382,7 @@ public class InterProceduralAnalysis
 	 * @param vars  Map of variables eligible for propagation.
 	 * @throws HopsException  If a HopsException occurs.
 	 */
-	private void propagateScalarsAcrossDAG(ArrayList<Hop> roots, LocalVariableMap vars)
+	private static void propagateScalarsAcrossDAG(ArrayList<Hop> roots, LocalVariableMap vars)
 		throws HopsException
 	{
 		for (Hop hop : roots) {
@@ -382,7 +394,7 @@ public class InterProceduralAnalysis
 		}
 	}
 
-	private void propagateStatisticsAcrossPredicateDAG( Hop root, LocalVariableMap vars ) 
+	private static void propagateStatisticsAcrossPredicateDAG( Hop root, LocalVariableMap vars ) 
 		throws HopsException
 	{
 		if( root == null )
@@ -407,7 +419,7 @@ public class InterProceduralAnalysis
 	 * @param vars  Map of variables eligible for propagation.
 	 * @throws HopsException  If a HopsException occurs.
 	 */
-	private void propagateStatisticsAcrossDAG( ArrayList<Hop> roots, LocalVariableMap vars )
+	private static void propagateStatisticsAcrossDAG( ArrayList<Hop> roots, LocalVariableMap vars )
 		throws HopsException
 	{
 		if( roots == null )
@@ -473,7 +485,7 @@ public class InterProceduralAnalysis
 		{
 			//maintain counters and investigate functions if not seen so far
 			FunctionOp fop = (FunctionOp) hop;
-			String fkey = DMLProgram.constructFunctionKey(fop.getFunctionNamespace(), fop.getFunctionName());
+			String fkey = fop.getFunctionKey();
 			
 			if( fop.getFunctionType() == FunctionType.DML )
 			{
@@ -522,12 +534,12 @@ public class InterProceduralAnalysis
 		hop.setVisited();
 	}
 	
-	private void populateLocalVariableMapForFunctionCall( FunctionStatement fstmt, FunctionOp fop, LocalVariableMap callvars, LocalVariableMap vars, FunctionCallSizeInfo fcallSizes ) 
+	private static void populateLocalVariableMapForFunctionCall( FunctionStatement fstmt, FunctionOp fop, LocalVariableMap callvars, LocalVariableMap vars, FunctionCallSizeInfo fcallSizes ) 
 		throws HopsException
 	{
 		ArrayList<DataIdentifier> inputVars = fstmt.getInputParams();
 		ArrayList<Hop> inputOps = fop.getInput();
-		String fkey = DMLProgram.constructFunctionKey(fop.getFunctionNamespace(), fop.getFunctionName());
+		String fkey = fop.getFunctionKey();
 		
 		for( int i=0; i<inputVars.size(); i++ )
 		{
@@ -542,7 +554,7 @@ public class InterProceduralAnalysis
 				MatrixCharacteristics mc = new MatrixCharacteristics( input.getDim1(), input.getDim2(), 
 					ConfigurationManager.getBlocksize(), ConfigurationManager.getBlocksize(),
 					fcallSizes.isSafeNnz(fkey, i)?input.getNnz():-1 );
-				MatrixFormatMetaData meta = new MatrixFormatMetaData(mc,null,null);
+				MetaDataFormat meta = new MetaDataFormat(mc,null,null);
 				mo.setMetaData(meta);	
 				vars.put(dat.getName(), mo);	
 			}
@@ -582,12 +594,12 @@ public class InterProceduralAnalysis
 	 *                      calling program's variable map.
 	 * @throws HopsException  If a HopsException occurs.
 	 */
-	private void extractFunctionCallReturnStatistics( FunctionStatement fstmt, FunctionOp fop, LocalVariableMap tmpVars, LocalVariableMap callVars, boolean overwrite ) 
+	private static void extractFunctionCallReturnStatistics( FunctionStatement fstmt, FunctionOp fop, LocalVariableMap tmpVars, LocalVariableMap callVars, boolean overwrite ) 
 		throws HopsException
 	{
 		ArrayList<DataIdentifier> foutputOps = fstmt.getOutputParams();
 		String[] outputVars = fop.getOutputVariableNames();
-		String fkey = DMLProgram.constructFunctionKey(fop.getFunctionNamespace(), fop.getFunctionName());
+		String fkey = fop.getFunctionKey();
 		
 		try
 		{
@@ -626,8 +638,9 @@ public class InterProceduralAnalysis
 						{
 							MatrixObject moOut = (MatrixObject)dat;
 							MatrixCharacteristics mc = moOut.getMatrixCharacteristics();
-							if( OptimizerUtils.estimateSizeExactSparsity(mc.getRows(), mc.getCols(), (mc.getNonZeros()>0)?((double)mc.getNonZeros())/mc.getRows()/mc.getCols():1.0)	
-							    < OptimizerUtils.estimateSize(moIn.getNumRows(), moIn.getNumColumns()) )
+							if( OptimizerUtils.estimateSizeExactSparsity(mc.getRows(), mc.getCols(), (mc.getNonZeros()>0)?
+								OptimizerUtils.getSparsity(mc):1.0) 
+								< OptimizerUtils.estimateSize(moIn.getNumRows(), moIn.getNumColumns()) )
 							{
 								//update statistics if necessary
 								mc.setDimension(moIn.getNumRows(), moIn.getNumColumns());
@@ -645,12 +658,12 @@ public class InterProceduralAnalysis
 		}
 	}
 	
-	private void extractFunctionCallUnknownReturnStatistics( FunctionStatement fstmt, FunctionOp fop, LocalVariableMap callVars ) 
+	private static void extractFunctionCallUnknownReturnStatistics( FunctionStatement fstmt, FunctionOp fop, LocalVariableMap callVars ) 
 		throws HopsException
 	{
 		ArrayList<DataIdentifier> foutputOps = fstmt.getOutputParams();
 		String[] outputVars = fop.getOutputVariableNames();
-		String fkey = DMLProgram.constructFunctionKey(fop.getFunctionNamespace(), fop.getFunctionName());
+		String fkey = fop.getFunctionKey();
 		
 		try
 		{
@@ -661,7 +674,7 @@ public class InterProceduralAnalysis
 				
 				if( di.getDataType()==DataType.MATRIX )
 				{
-					MatrixObject moOut = createOutputMatrix(-1, -1, -1);	
+					MatrixObject moOut = createOutputMatrix(-1, -1, -1);
 					callVars.put(pvarname, moOut);
 				}
 			}
@@ -672,21 +685,21 @@ public class InterProceduralAnalysis
 		}
 	}
 	
-	private void extractFunctionCallEquivalentReturnStatistics( FunctionStatement fstmt, FunctionOp fop, LocalVariableMap callVars ) 
+	private static void extractFunctionCallEquivalentReturnStatistics( FunctionStatement fstmt, FunctionOp fop, LocalVariableMap callVars ) 
 		throws HopsException
 	{
-		String fkey = DMLProgram.constructFunctionKey(fop.getFunctionNamespace(), fop.getFunctionName());
 		try {
 			Hop input = fop.getInput().get(0);
 			MatrixObject moOut = createOutputMatrix(input.getDim1(), input.getDim2(), -1);	
 			callVars.put(fop.getOutputVariableNames()[0], moOut);
 		}
 		catch( Exception ex ) {
-			throw new HopsException( "Failed to extract output statistics for unary function "+fkey+".", ex);
+			throw new HopsException( "Failed to extract output statistics "
+				+ "for unary function "+fop.getFunctionKey()+".", ex);
 		}
 	}
 	
-	private void extractExternalFunctionCallReturnStatistics( ExternalFunctionStatement fstmt, FunctionOp fop, LocalVariableMap callVars ) 
+	private static void extractExternalFunctionCallReturnStatistics( ExternalFunctionStatement fstmt, FunctionOp fop, LocalVariableMap callVars ) 
 		throws HopsException
 	{
 		String className = fstmt.getOtherParams().get(ExternalFunctionStatement.CLASS_NAME);
@@ -713,11 +726,11 @@ public class InterProceduralAnalysis
 		}
 	}
 	
-	private MatrixObject createOutputMatrix( long dim1, long dim2, long nnz ) {
+	private static MatrixObject createOutputMatrix( long dim1, long dim2, long nnz ) {
 		MatrixObject moOut = new MatrixObject(ValueType.DOUBLE, null);
 		MatrixCharacteristics mc = new MatrixCharacteristics( dim1, dim2,
 				ConfigurationManager.getBlocksize(), ConfigurationManager.getBlocksize(), nnz);
-		MatrixFormatMetaData meta = new MatrixFormatMetaData(mc,null,null);
+		MetaDataFormat meta = new MetaDataFormat(mc,null,null);
 		moOut.setMetaData(meta);
 		
 		return moOut;
