@@ -23,10 +23,13 @@ import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.HashMap;
 import java.util.HashSet;
+import java.util.Set;
+import java.util.stream.Collectors;
 
 import org.antlr.v4.runtime.ParserRuleContext;
 import org.apache.sysml.hops.Hop.ParamBuiltinOp;
 import org.apache.sysml.parser.LanguageException.LanguageErrorCodes;
+import org.apache.sysml.runtime.util.UtilFunctions;
 import org.apache.wink.json4j.JSONObject;
 
 
@@ -476,6 +479,15 @@ public class ParameterizedBuiltinFunctionExpression extends DataIdentifier
 	}
 
 	private void validateRemoveEmpty(DataIdentifier output, boolean conditional) throws LanguageException {
+		
+		//check for invalid parameters
+		Set<String> valid = UtilFunctions.asSet("target", "margin", "select", "empty.return");
+		Set<String> invalid = _varParams.keySet().stream()
+			.filter(k -> !valid.contains(k)).collect(Collectors.toSet());
+		if( !invalid.isEmpty() )
+			raiseValidateError("Invalid parameters for removeEmpty: "
+				+ Arrays.toString(invalid.toArray(new String[0])), false);
+		
 		//check existence and correctness of arguments
 		Expression target = getVarParam("target");
 		if( target==null ) {
@@ -498,11 +510,18 @@ public class ParameterizedBuiltinFunctionExpression extends DataIdentifier
 			raiseValidateError("Index matrix 'select' is of type '"+select.getOutput().getDataType()+"'. Please specify the select matrix.", conditional, LanguageErrorCodes.INVALID_PARAMETERS);
 		}
 		
+		Expression empty = getVarParam("empty.return");
+		if( empty!=null && (!empty.getOutput().getDataType().isScalar() || empty.getOutput().getValueType() != ValueType.BOOLEAN) ){
+			raiseValidateError("Boolean parameter 'empty.return' is of type "+empty.getOutput().getDataType()
+				+"["+empty.getOutput().getValueType()+"].", conditional, LanguageErrorCodes.INVALID_PARAMETERS);
+		}
+		if( empty == null ) //default handling
+			_varParams.put("empty.return", new BooleanIdentifier(true));
+		
 		// Output is a matrix with unknown dims
 		output.setDataType(DataType.MATRIX);
 		output.setValueType(ValueType.DOUBLE);
 		output.setDimensions(-1, -1);
-		
 	}
 	
 	private void validateGroupedAgg(DataIdentifier output, boolean conditional) 
