@@ -21,6 +21,7 @@ package org.apache.sysml.parser;
 
 import java.io.IOException;
 import java.util.ArrayList;
+import java.util.Arrays;
 import java.util.HashMap;
 import java.util.List;
 
@@ -31,6 +32,7 @@ import org.apache.sysml.conf.ConfigurationManager;
 import org.apache.sysml.hops.Hop;
 import org.apache.sysml.hops.HopsException;
 import org.apache.sysml.hops.recompile.Recompiler;
+import org.apache.sysml.hops.rewrite.RewriteSplitDagDataDependentOperators;
 import org.apache.sysml.lops.Lop;
 import org.apache.sysml.parser.Expression.DataType;
 import org.apache.sysml.parser.Expression.FormatType;
@@ -60,16 +62,22 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 
 	public StatementBlock() {
 		_dmlProg = null;
-		_statements = new ArrayList<Statement>();
+		_statements = new ArrayList<>();
 		_read = new VariableSet();
 		_updated = new VariableSet();
 		_gen = new VariableSet();
 		_kill = new VariableSet();
 		_warnSet = new VariableSet();
 		_initialized = true;
-		_constVarsIn = new HashMap<String,ConstIdentifier>();
-		_constVarsOut = new HashMap<String,ConstIdentifier>();
-		_updateInPlaceVars = new ArrayList<String>();
+		_constVarsIn = new HashMap<>();
+		_constVarsOut = new HashMap<>();
+		_updateInPlaceVars = new ArrayList<>();
+	}
+	
+	public StatementBlock(StatementBlock sb) {
+		this();
+		setParseInfo(sb);
+		_dmlProg = sb._dmlProg;
 	}
 
 	public void setDMLProg(DMLProgram dmlProg){
@@ -114,8 +122,7 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 		return _statements.get(i);
 	}
 
-	public ArrayList<Statement> getStatements()
-	{
+	public ArrayList<Statement> getStatements() {
 		return _statements;
 	}
 
@@ -124,7 +131,7 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 		_statements = s;
 	}
 
-	public ArrayList<Hop> get_hops() throws HopsException {
+	public ArrayList<Hop> getHops() {
 		return _hops;
 	}
 
@@ -132,7 +139,7 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 		return _lops;
 	}
 
-	public void set_hops(ArrayList<Hop> hops) {
+	public void setHops(ArrayList<Hop> hops) {
 		_hops = hops;
 	}
 
@@ -156,6 +163,9 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 		return _splitDag;
 	}
 
+	private boolean isMergeablePrintStatement(Statement stmt) {
+		return ( stmt instanceof PrintStatement && (((PrintStatement)stmt).getType() == PRINTTYPE.STOP || ((PrintStatement)stmt).getType() == PRINTTYPE.ASSERT) );
+	}
 
     public boolean isMergeableFunctionCallBlock(DMLProgram dmlProg) throws LanguageException{
 
@@ -168,7 +178,7 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 
 		// Check whether targetIndex block is: control stmt block or stmt block for un-mergable function call
 		if (   stmt instanceof WhileStatement || stmt instanceof IfStatement || stmt instanceof ForStatement
-			|| stmt instanceof FunctionStatement || ( stmt instanceof PrintStatement && ((PrintStatement)stmt).getType() == PRINTTYPE.STOP )/*|| stmt instanceof ELStatement*/ )
+			|| stmt instanceof FunctionStatement || isMergeablePrintStatement(stmt) /*|| stmt instanceof ELStatement*/ )
 		{
 			return false;
 		}
@@ -322,8 +332,7 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 			}
 		}
 
-		ArrayList<StatementBlock> result = new ArrayList<StatementBlock>();
-
+		ArrayList<StatementBlock> result = new ArrayList<>();
 		StatementBlock currentBlock = null;
 
 		for (int i = 0; i < body.size(); i++){
@@ -350,6 +359,7 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 		return result;
 	}
 
+	@Override
 	public String toString(){
 		StringBuilder sb = new StringBuilder();
 		sb.append("statements\n");
@@ -367,13 +377,10 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 	}
 
 	public static ArrayList<StatementBlock> mergeStatementBlocks(ArrayList<StatementBlock> sb){
+		if (sb == null || sb.isEmpty())
+			return new ArrayList<>();
 
-		ArrayList<StatementBlock> result = new ArrayList<StatementBlock>();
-
-		if (sb == null || sb.isEmpty()) {
-			return new ArrayList<StatementBlock>();
-		}
-
+		ArrayList<StatementBlock> result = new ArrayList<>();
 		StatementBlock currentBlock = null;
 
 		for (int i = 0; i < sb.size(); i++){
@@ -400,11 +407,175 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 		return result;
 
 	}
+	
+	public static List<StatementBlock> rHoistFunctionCallsFromExpressions(StatementBlock current) {
+		if (current instanceof FunctionStatementBlock) {
+			FunctionStatementBlock fsb = (FunctionStatementBlock)current;
+			FunctionStatement fstmt = (FunctionStatement)fsb.getStatement(0);
+			ArrayList<StatementBlock> tmp = new ArrayList<>();
+			for (StatementBlock sb : fstmt.getBody())
+				tmp.addAll(rHoistFunctionCallsFromExpressions(sb));
+			fstmt.setBody(tmp);
+		}
+		else if (current instanceof WhileStatementBlock) {
+			WhileStatementBlock wsb = (WhileStatementBlock) current;
+			WhileStatement wstmt = (WhileStatement)wsb.getStatement(0);
+			//TODO handle predicates
+			ArrayList<StatementBlock> tmp = new ArrayList<>();
+			for (StatementBlock sb : wstmt.getBody())
+				tmp.addAll(rHoistFunctionCallsFromExpressions(sb));
+			wstmt.setBody(tmp);
+		}
+		else if (current instanceof IfStatementBlock) {
+			IfStatementBlock isb = (IfStatementBlock) current;
+			IfStatement istmt = (IfStatement)isb.getStatement(0);
+			//TODO handle predicates
+			ArrayList<StatementBlock> tmp = new ArrayList<>();
+			for (StatementBlock sb : istmt.getIfBody())
+				tmp.addAll(rHoistFunctionCallsFromExpressions(sb));
+			istmt.setIfBody(tmp);
+			if( istmt.getElseBody() != null && !istmt.getElseBody().isEmpty() ) {
+				ArrayList<StatementBlock> tmp2 = new ArrayList<>();
+				for (StatementBlock sb : istmt.getElseBody())
+					tmp2.addAll(rHoistFunctionCallsFromExpressions(sb));
+				istmt.setElseBody(tmp2);
+			}
+		}
+		else if (current instanceof ForStatementBlock) { //incl parfor
+			ForStatementBlock fsb = (ForStatementBlock) current;
+			ForStatement fstmt = (ForStatement)fsb.getStatement(0);
+			//TODO handle predicates
+			ArrayList<StatementBlock> tmp = new ArrayList<>();
+			for (StatementBlock sb : fstmt.getBody())
+				tmp.addAll(rHoistFunctionCallsFromExpressions(sb));
+			fstmt.setBody(tmp);
+		}
+		else { //generic (last-level)
+			ArrayList<Statement> tmp = new ArrayList<>();
+			for(Statement stmt : current.getStatements())
+				tmp.addAll(rHoistFunctionCallsFromExpressions(stmt));
+			if( current.getStatements().size() != tmp.size() )
+				return createStatementBlocks(current, tmp);
+		}
+		return Arrays.asList(current);
+	}
 
-
+	public static List<Statement> rHoistFunctionCallsFromExpressions(Statement stmt) {
+		ArrayList<Statement> tmp = new ArrayList<>();
+		if( stmt instanceof AssignmentStatement ) {
+			AssignmentStatement astmt = (AssignmentStatement)stmt;
+			boolean ix = (astmt.getTargetList().get(0) instanceof IndexedIdentifier);
+			rHoistFunctionCallsFromExpressions(astmt.getSource(), !ix, tmp);
+			if( ix && astmt.getSource() instanceof FunctionCallIdentifier ) {
+				AssignmentStatement lstmt = (AssignmentStatement) tmp.get(tmp.size()-1);
+				astmt.setSource(copy(lstmt.getTarget()));
+			}
+		}
+		else if( stmt instanceof MultiAssignmentStatement ) {
+			MultiAssignmentStatement mstmt = (MultiAssignmentStatement)stmt;
+			rHoistFunctionCallsFromExpressions(mstmt.getSource(), true, tmp);
+		}
+		else if( stmt instanceof PrintStatement ) {
+			PrintStatement pstmt = (PrintStatement)stmt;
+			for(int i=0; i<pstmt.expressions.size(); i++) {
+				Expression lexpr = pstmt.getExpressions().get(i);
+				rHoistFunctionCallsFromExpressions(lexpr, false, tmp);
+				if( lexpr instanceof FunctionCallIdentifier ) {
+					AssignmentStatement lstmt = (AssignmentStatement) tmp.get(tmp.size()-1);
+					pstmt.getExpressions().set(i, copy(lstmt.getTarget()));
+				}
+			}
+		}
+		
+		//most statements will be returned unchanged, while expressions with
+		//function calls are split into potentially many statements
+		List<Statement> ret = tmp.isEmpty() ? Arrays.asList(stmt) : tmp;
+		if( !tmp.isEmpty() ) {
+			for( Statement ltmp : tmp )
+				ltmp.setParseInfo(stmt);
+			tmp.add(stmt);
+		}
+		return ret;
+	}
+	
+	public static Expression rHoistFunctionCallsFromExpressions(Expression expr, boolean root, ArrayList<Statement> tmp) {
+		if( expr == null || expr instanceof ConstIdentifier )
+			return expr; //do nothing
+		if( expr instanceof BinaryExpression ) {
+			BinaryExpression lexpr = (BinaryExpression) expr;
+			lexpr.setLeft(rHoistFunctionCallsFromExpressions(lexpr.getLeft(), false, tmp));
+			lexpr.setRight(rHoistFunctionCallsFromExpressions(lexpr.getRight(), false, tmp));
+		}
+		else if( expr instanceof RelationalExpression ) {
+			RelationalExpression lexpr = (RelationalExpression) expr;
+			lexpr.setLeft(rHoistFunctionCallsFromExpressions(lexpr.getLeft(), false, tmp));
+			lexpr.setRight(rHoistFunctionCallsFromExpressions(lexpr.getRight(), false, tmp));
+		}
+		else if( expr instanceof BooleanExpression ) {
+			BooleanExpression lexpr = (BooleanExpression) expr;
+			lexpr.setLeft(rHoistFunctionCallsFromExpressions(lexpr.getLeft(), false, tmp));
+			lexpr.setRight(rHoistFunctionCallsFromExpressions(lexpr.getRight(), false, tmp));
+		}
+		else if( expr instanceof BuiltinFunctionExpression ) {
+			BuiltinFunctionExpression lexpr = (BuiltinFunctionExpression) expr;
+			Expression[] clexpr = lexpr.getAllExpr();
+			for( int i=0; i<clexpr.length; i++ )
+				clexpr[i] = rHoistFunctionCallsFromExpressions(clexpr[i], false, tmp);
+		}
+		else if( expr instanceof ParameterizedBuiltinFunctionExpression ) {
+			ParameterizedBuiltinFunctionExpression lexpr = (ParameterizedBuiltinFunctionExpression) expr;
+			HashMap<String, Expression> clexpr = lexpr.getVarParams();
+			for( String key : clexpr.keySet() )
+				clexpr.put(key, rHoistFunctionCallsFromExpressions(clexpr.get(key), false, tmp));
+		}
+		else if( expr instanceof DataExpression ) {
+			DataExpression lexpr = (DataExpression) expr;
+			HashMap<String, Expression> clexpr = lexpr.getVarParams();
+			for( String key : clexpr.keySet() )
+				clexpr.put(key, rHoistFunctionCallsFromExpressions(clexpr.get(key), false, tmp));
+		}
+		else if( expr instanceof FunctionCallIdentifier ) {
+			FunctionCallIdentifier fexpr = (FunctionCallIdentifier) expr;
+			for( ParameterExpression pexpr : fexpr.getParamExprs() )
+				pexpr.setExpr(rHoistFunctionCallsFromExpressions(pexpr.getExpr(), false, tmp));
+			if( !root ) { //core hoisting
+				String varname = RewriteSplitDagDataDependentOperators.createCutVarName(true);
+				DataIdentifier di = new DataIdentifier(varname);
+				di.setDataType(fexpr.getDataType());
+				di.setValueType(fexpr.getValueType());
+				tmp.add(new AssignmentStatement(di, fexpr, di));
+				return di;
+			}
+		}
+		//note: all remaining expressions data identifiers remain unchanged
+		return expr;
+	}
+	
+	private static DataIdentifier copy(DataIdentifier di) {
+		return new DataIdentifier(di);
+	}
+	
+	private static List<StatementBlock> createStatementBlocks(StatementBlock sb, List<Statement> stmts) {
+		List<StatementBlock> ret = new ArrayList<StatementBlock>();
+		StatementBlock current = new StatementBlock(sb);
+		for(Statement stmt : stmts) {
+			current.addStatement(stmt);
+			//cut the statement block after the current function
+			if( stmt instanceof AssignmentStatement
+				&& ((AssignmentStatement)stmt).getSource()
+				instanceof FunctionCallIdentifier ) {
+				ret.add(current);
+				current = new StatementBlock(sb);
+			}
+		}
+		if( current.getNumStatements() > 0 )
+			ret.add(current);
+		return ret;
+	}
+	
 	public ArrayList<Statement> rewriteFunctionCallStatements (DMLProgram dmlProg, ArrayList<Statement> statements) throws LanguageException {
 
-		ArrayList<Statement> newStatements = new ArrayList<Statement>();
+		ArrayList<Statement> newStatements = new ArrayList<>();
 		for (Statement current : statements){
 			if (isRewritableFunctionCall(current, dmlProg)){
 
@@ -545,24 +716,19 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 		throws LanguageException, ParseException, IOException
 	{
 		_constVarsIn.putAll(constVars);
-		HashMap<String, ConstIdentifier> currConstVars = new HashMap<String,ConstIdentifier>();
-		currConstVars.putAll(constVars);
-
 		_statements = rewriteFunctionCallStatements(dmlProg, _statements);
 		_dmlProg = dmlProg;
-
-		for (Statement current : _statements){
-
-			if (current instanceof OutputStatement){
+		
+		HashMap<String, ConstIdentifier> currConstVars = new HashMap<>(constVars);
+		for (Statement current : _statements) {
+			if (current instanceof OutputStatement) {
 				OutputStatement os = (OutputStatement)current;
-
 				// validate variable being written by output statement exists
 				DataIdentifier target = (DataIdentifier)os.getIdentifier();
 				if (ids.getVariable(target.getName()) == null) {
 					//undefined variables are always treated unconditionally as error in order to prevent common script-level bugs
 					raiseValidateError("Undefined Variable (" + target.getName() + ") used in statement", false, LanguageErrorCodes.INVALID_PARAMETERS);
 				}
-
 				if ( ids.getVariable(target.getName()).getDataType() == DataType.SCALAR) {
 					boolean paramsOkay = true;
 					for (String key : os.getSource().getVarParams().keySet()){
@@ -577,197 +743,18 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 				Expression source = os.getSource();
 				source.setOutput(target);
 				source.validateExpression(ids.getVariables(), currConstVars, conditional);
-
 				setStatementFormatType(os, conditional);
 				target.setDimensionValueProperties(ids.getVariable(target.getName()));
 			}
-
 			else if (current instanceof AssignmentStatement){
-				AssignmentStatement as = (AssignmentStatement)current;
-				DataIdentifier target = as.getTarget();
-			 	Expression source = as.getSource();
-
-				if (source instanceof FunctionCallIdentifier) {
-					((FunctionCallIdentifier) source).validateExpression(dmlProg, ids.getVariables(),currConstVars, conditional);
-				}
-				else {
-					if( MLContextProxy.isActive() )
-						MLContextProxy.setAppropriateVarsForRead(source, target._name);
-
-					source.validateExpression(ids.getVariables(), currConstVars, conditional);
-				}
-
-				if (source instanceof DataExpression && ((DataExpression)source).getOpCode() == Expression.DataOp.READ)
-					setStatementFormatType(as, conditional);
-
-				// Handle const vars: (a) basic constant propagation, and (b) transitive constant propagation over assignments
-				if (target != null) {
-					currConstVars.remove(target.getName());
-					if(source instanceof ConstIdentifier && !(target instanceof IndexedIdentifier)){ //basic
-						currConstVars.put(target.getName(), (ConstIdentifier)source);
-					}
-					if( source instanceof DataIdentifier && !(target instanceof IndexedIdentifier) ){ //transitive
-						DataIdentifier diSource = (DataIdentifier) source;
-						if( currConstVars.containsKey(diSource.getName()) ){
-							currConstVars.put(target.getName(), currConstVars.get(diSource.getName()));
-						}
-					}
-				}
-
-				if (source instanceof BuiltinFunctionExpression){
-					BuiltinFunctionExpression bife = (BuiltinFunctionExpression)source;
-					if (   bife.getOpCode() == Expression.BuiltinFunctionOp.NROW
-						|| bife.getOpCode() == Expression.BuiltinFunctionOp.NCOL )
-					{
-						DataIdentifier id = (DataIdentifier)bife.getFirstExpr();
-						DataIdentifier currVal = ids.getVariable(id.getName());
-						if (currVal == null){
-							//undefined variables are always treated unconditionally as error in order to prevent common script-level bugs
-							bife.raiseValidateError("Undefined Variable (" + id.getName() + ") used in statement", false, LanguageErrorCodes.INVALID_PARAMETERS);
-						}
-						IntIdentifier intid = null;
-						if (bife.getOpCode() == Expression.BuiltinFunctionOp.NROW) {
-							intid = new IntIdentifier((currVal instanceof IndexedIdentifier)
-									? ((IndexedIdentifier) currVal).getOrigDim1() : currVal.getDim1(), bife);
-						} else {
-							intid = new IntIdentifier((currVal instanceof IndexedIdentifier)
-									? ((IndexedIdentifier) currVal).getOrigDim2() : currVal.getDim2(), bife);
-						}
-
-						// handle case when nrow / ncol called on variable with size unknown (dims == -1)
-						//	--> const prop NOT possible
-						if (intid.getValue() != -1){
-							currConstVars.put(target.getName(), intid);
-						}
-					}
-				}
-				if (target == null) {
-					// function has no return value
-				}
-				// CASE: target NOT indexed identifier
-				else if (!(target instanceof IndexedIdentifier)){
-					target.setProperties(source.getOutput());
-					if (source.getOutput() instanceof IndexedIdentifier){
-						target.setDimensions(source.getOutput().getDim1(), source.getOutput().getDim2());
-					}
-
-				}
-				// CASE: target is indexed identifier
-				else
-				{
-					// process the "target" being indexed
-					DataIdentifier targetAsSeen = ids.getVariable(target.getName());
-					if (targetAsSeen == null){
-						target.raiseValidateError("cannot assign value to indexed identifier " + target.toString() + " without first initializing " + target.getName(), conditional);
-					}
-					target.setProperties(targetAsSeen);
-
-					// process the expressions for the indexing
-					if ( ((IndexedIdentifier)target).getRowLowerBound() != null  )
-						((IndexedIdentifier)target).getRowLowerBound().validateExpression(ids.getVariables(), currConstVars, conditional);
-					if ( ((IndexedIdentifier)target).getRowUpperBound() != null  )
-						((IndexedIdentifier)target).getRowUpperBound().validateExpression(ids.getVariables(), currConstVars, conditional);
-					if ( ((IndexedIdentifier)target).getColLowerBound() != null  )
-						((IndexedIdentifier)target).getColLowerBound().validateExpression(ids.getVariables(), currConstVars, conditional);
-					if ( ((IndexedIdentifier)target).getColUpperBound() != null  )
-						((IndexedIdentifier)target).getColUpperBound().validateExpression(ids.getVariables(), currConstVars, conditional);
-
-					// validate that LHS indexed identifier is being assigned a matrix value
-//					if (source.getOutput().getDataType() != Expression.DataType.MATRIX){
-//						LOG.error(target.printErrorLocation() + "Indexed expression " + target.toString() + " can only be assigned matrix value");
-//						throw new LanguageException(target.printErrorLocation() + "Indexed expression " + target.toString() + " can only be assigned matrix value");
-//					}
-
-					// validate that size of LHS index ranges is being assigned:
-					//	(a) a matrix value of same size as LHS
-					//	(b) singleton value (semantics: initialize enitre submatrix with this value)
-					IndexPair targetSize = ((IndexedIdentifier)target).calculateIndexedDimensions(ids.getVariables(), currConstVars, conditional);
-
-					if (targetSize._row >= 1 && source.getOutput().getDim1() > 1 && targetSize._row != source.getOutput().getDim1()){
-						target.raiseValidateError("Dimension mismatch. Indexed expression " + target.toString() + " can only be assigned matrix with dimensions "
-								+ targetSize._row + " rows and " + targetSize._col + " cols. Attempted to assign matrix with dimensions "
-								+ source.getOutput().getDim1() + " rows and " + source.getOutput().getDim2() + " cols ", conditional);
-					}
-
-					if (targetSize._col >= 1 && source.getOutput().getDim2() > 1 && targetSize._col != source.getOutput().getDim2()){
-						target.raiseValidateError("Dimension mismatch. Indexed expression " + target.toString() + " can only be assigned matrix with dimensions "
-								+ targetSize._row + " rows and " + targetSize._col + " cols. Attempted to assign matrix with dimensions "
-								+ source.getOutput().getDim1() + " rows and " + source.getOutput().getDim2() + " cols ", conditional);
-					}
-
-					((IndexedIdentifier)target).setDimensions(targetSize._row, targetSize._col);
-				}
-
-				if (target != null) {
-					ids.addVariable(target.getName(), target);
-				}
-
+				validateAssignmentStatement(current, dmlProg, ids, currConstVars, conditional);
 			}
-
 			else if (current instanceof MultiAssignmentStatement){
-				MultiAssignmentStatement mas = (MultiAssignmentStatement) current;
-				ArrayList<DataIdentifier> targetList = mas.getTargetList();
-
-				// perform validation of source expression
-				Expression source = mas.getSource();
-				/*
-				 * MultiAssignmentStatments currently supports only External,
-				 * User-defined, and Multi-return Builtin function expressions
-				 */
-				if (!(source instanceof DataIdentifier)
-						|| (source instanceof DataIdentifier && !((DataIdentifier)source).multipleReturns()) ) {
-				//if (!(source instanceof FunctionCallIdentifier) ) {
-						//|| !(source instanceof BuiltinFunctionExpression && ((BuiltinFunctionExpression)source).isMultiReturnBuiltinFunction()) ){
-					source.raiseValidateError("can only use user-defined functions with multi-assignment statement", conditional);
-				}
-
-				if ( source instanceof FunctionCallIdentifier) {
-					FunctionCallIdentifier fci = (FunctionCallIdentifier)source;
-					fci.validateExpression(dmlProg, ids.getVariables(), currConstVars, conditional);
-				}
-				else if ( (source instanceof BuiltinFunctionExpression || source instanceof ParameterizedBuiltinFunctionExpression)
-						&& ((DataIdentifier)source).multipleReturns()) {
-					source.validateExpression(mas, ids.getVariables(), currConstVars, conditional);
-				}
-				else
-					throw new LanguageException("Unexpected error.");
-
-
-				if ( source instanceof FunctionCallIdentifier ) {
-					for (int j =0; j< targetList.size(); j++){
-
-						DataIdentifier target = targetList.get(j);
-							// set target properties (based on type info in function call statement return params)
-							FunctionCallIdentifier fci = (FunctionCallIdentifier)source;
-							FunctionStatement fstmt = (FunctionStatement)_dmlProg.getFunctionStatementBlock(fci.getNamespace(), fci.getName()).getStatement(0);
-							if (fstmt == null){
-								fci.raiseValidateError(" function " + fci.getName() + " is undefined in namespace " + fci.getNamespace(), conditional);
-							}
-							if (!(target instanceof IndexedIdentifier)){
-								target.setProperties(fstmt.getOutputParams().get(j));
-							}
-							else{
-								DataIdentifier targetAsSeen = ids.getVariable(target.getName());
-								if (targetAsSeen == null){
-									raiseValidateError(target.printErrorLocation() + "cannot assign value to indexed identifier " + target.toString() + " without first initializing " + target.getName(), conditional);
-								}
-								target.setProperties(targetAsSeen);
-							}
-							ids.addVariable(target.getName(), target);
-					}
-				}
-				else if ( source instanceof BuiltinFunctionExpression || source instanceof ParameterizedBuiltinFunctionExpression ) {
-					Identifier[] outputs = source.getOutputs();
-					for (int j=0; j < targetList.size(); j++) {
-						ids.addVariable(targetList.get(j).getName(), (DataIdentifier)outputs[j]);
-					}
-				}
+				validateMultiAssignmentStatement(current, dmlProg, ids, currConstVars, conditional);
 			}
-
 			else if(current instanceof ForStatement || current instanceof IfStatement || current instanceof WhileStatement ){
 				raiseValidateError("control statement (WhileStatement, IfStatement, ForStatement) should not be in generic statement block. Likely a parsing error", conditional);
 			}
-
 			else if (current instanceof PrintStatement) {
 				PrintStatement pstmt = (PrintStatement) current;
 				List<Expression> expressions = pstmt.getExpressions();
@@ -782,22 +769,192 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 					}
 				}
 			}
-
 			// no work to perform for PathStatement or ImportStatement
 			else if (current instanceof PathStatement){}
 			else if (current instanceof ImportStatement){}
-
-
 			else {
 				raiseValidateError("cannot process statement of type " + current.getClass().getSimpleName(), conditional);
 			}
-
-		} // end for (Statement current : _statements){
+		}
 		_constVarsOut.putAll(currConstVars);
 		return ids;
-
 	}
-
+	
+	private void validateAssignmentStatement(Statement current, DMLProgram dmlProg, 
+		VariableSet ids, HashMap<String, ConstIdentifier> currConstVars, boolean conditional) 
+			throws LanguageException, IOException, ParseException 
+	{
+		AssignmentStatement as = (AssignmentStatement)current;
+		DataIdentifier target = as.getTarget();
+		Expression source = as.getSource();
+		
+		if (source instanceof FunctionCallIdentifier) {
+			((FunctionCallIdentifier) source).validateExpression(
+				dmlProg, ids.getVariables(),currConstVars, conditional);
+		}
+		else { //all builtin functions and expressions
+			if( target == null  )
+				raiseValidateError("Missing variable assignment.", false);
+			
+			if( MLContextProxy.isActive() )
+				MLContextProxy.setAppropriateVarsForRead(source, target._name);
+			
+			source.validateExpression(ids.getVariables(), currConstVars, conditional);
+		}
+		
+		if (source instanceof DataExpression && ((DataExpression)source).getOpCode() == Expression.DataOp.READ)
+			setStatementFormatType(as, conditional);
+		
+		// Handle const vars: (a) basic constant propagation, and (b) transitive constant propagation over assignments
+		if (target != null) {
+			currConstVars.remove(target.getName());
+			if(source instanceof ConstIdentifier && !(target instanceof IndexedIdentifier)){ //basic
+				currConstVars.put(target.getName(), (ConstIdentifier)source);
+			}
+			if( source instanceof DataIdentifier && !(target instanceof IndexedIdentifier) ){ //transitive
+				DataIdentifier diSource = (DataIdentifier) source;
+				if( currConstVars.containsKey(diSource.getName()) ){
+					currConstVars.put(target.getName(), currConstVars.get(diSource.getName()));
+				}
+			}
+		}
+		
+		if (source instanceof BuiltinFunctionExpression){
+			BuiltinFunctionExpression bife = (BuiltinFunctionExpression)source;
+			if (   bife.getOpCode() == Expression.BuiltinFunctionOp.NROW
+				|| bife.getOpCode() == Expression.BuiltinFunctionOp.NCOL )
+			{
+				DataIdentifier id = (DataIdentifier)bife.getFirstExpr();
+				DataIdentifier currVal = ids.getVariable(id.getName());
+				if (currVal == null){
+					//undefined variables are always treated unconditionally as error in order to prevent common script-level bugs
+					bife.raiseValidateError("Undefined Variable (" + id.getName() + ") used in statement", false, LanguageErrorCodes.INVALID_PARAMETERS);
+				}
+				IntIdentifier intid = null;
+				if (bife.getOpCode() == Expression.BuiltinFunctionOp.NROW) {
+					intid = new IntIdentifier((currVal instanceof IndexedIdentifier)
+							? ((IndexedIdentifier) currVal).getOrigDim1() : currVal.getDim1(), bife);
+				} else {
+					intid = new IntIdentifier((currVal instanceof IndexedIdentifier)
+							? ((IndexedIdentifier) currVal).getOrigDim2() : currVal.getDim2(), bife);
+				}
+				
+				// handle case when nrow / ncol called on variable with size unknown (dims == -1)
+				//	--> const prop NOT possible
+				if (intid.getValue() != -1)
+					currConstVars.put(target.getName(), intid);
+			}
+		}
+		if (target == null) {
+			// function has no return value
+		}
+		// CASE: target NOT indexed identifier
+		else if (!(target instanceof IndexedIdentifier)){
+			target.setProperties(source.getOutput());
+			if (source.getOutput() instanceof IndexedIdentifier)
+				target.setDimensions(source.getOutput().getDim1(), source.getOutput().getDim2());
+		}
+		// CASE: target is indexed identifier
+		else
+		{
+			// process the "target" being indexed
+			DataIdentifier targetAsSeen = ids.getVariable(target.getName());
+			if (targetAsSeen == null){
+				target.raiseValidateError("cannot assign value to indexed identifier " + target.toString() + " without first initializing " + target.getName(), conditional);
+			}
+			target.setProperties(targetAsSeen);
+			
+			// process the expressions for the indexing
+			if ( ((IndexedIdentifier)target).getRowLowerBound() != null  )
+				((IndexedIdentifier)target).getRowLowerBound().validateExpression(ids.getVariables(), currConstVars, conditional);
+			if ( ((IndexedIdentifier)target).getRowUpperBound() != null  )
+				((IndexedIdentifier)target).getRowUpperBound().validateExpression(ids.getVariables(), currConstVars, conditional);
+			if ( ((IndexedIdentifier)target).getColLowerBound() != null  )
+				((IndexedIdentifier)target).getColLowerBound().validateExpression(ids.getVariables(), currConstVars, conditional);
+			if ( ((IndexedIdentifier)target).getColUpperBound() != null  )
+				((IndexedIdentifier)target).getColUpperBound().validateExpression(ids.getVariables(), currConstVars, conditional);
+			
+			// validate that size of LHS index ranges is being assigned:
+			//	(a) a matrix value of same size as LHS
+			//	(b) singleton value (semantics: initialize enitre submatrix with this value)
+			IndexPair targetSize = ((IndexedIdentifier)target).calculateIndexedDimensions(ids.getVariables(), currConstVars, conditional);
+			
+			if (targetSize._row >= 1 && source.getOutput().getDim1() > 1 && targetSize._row != source.getOutput().getDim1()){
+				target.raiseValidateError("Dimension mismatch. Indexed expression " + target.toString() + " can only be assigned matrix with dimensions "
+						+ targetSize._row + " rows and " + targetSize._col + " cols. Attempted to assign matrix with dimensions "
+						+ source.getOutput().getDim1() + " rows and " + source.getOutput().getDim2() + " cols ", conditional);
+			}
+			
+			if (targetSize._col >= 1 && source.getOutput().getDim2() > 1 && targetSize._col != source.getOutput().getDim2()){
+				target.raiseValidateError("Dimension mismatch. Indexed expression " + target.toString() + " can only be assigned matrix with dimensions "
+						+ targetSize._row + " rows and " + targetSize._col + " cols. Attempted to assign matrix with dimensions "
+						+ source.getOutput().getDim1() + " rows and " + source.getOutput().getDim2() + " cols ", conditional);
+			}
+			((IndexedIdentifier)target).setDimensions(targetSize._row, targetSize._col);
+		}
+		
+		if (target != null)
+			ids.addVariable(target.getName(), target);
+	}
+	
+	private void validateMultiAssignmentStatement(Statement current, DMLProgram dmlProg, 
+		VariableSet ids, HashMap<String, ConstIdentifier> currConstVars, boolean conditional) 
+			throws LanguageException, IOException 
+	{
+		MultiAssignmentStatement mas = (MultiAssignmentStatement) current;
+		ArrayList<DataIdentifier> targetList = mas.getTargetList();
+		Expression source = mas.getSource();
+		
+		//MultiAssignmentStatments currently supports only External,
+		//User-defined, and Multi-return Builtin function expressions
+		if (!(source instanceof DataIdentifier)
+				|| (source instanceof DataIdentifier && !((DataIdentifier)source).multipleReturns()) ) {
+			source.raiseValidateError("can only use user-defined functions with multi-assignment statement", conditional);
+		}
+		if ( source instanceof FunctionCallIdentifier) {
+			FunctionCallIdentifier fci = (FunctionCallIdentifier)source;
+			fci.validateExpression(dmlProg, ids.getVariables(), currConstVars, conditional);
+		}
+		else if ( (source instanceof BuiltinFunctionExpression || source instanceof ParameterizedBuiltinFunctionExpression)
+				&& ((DataIdentifier)source).multipleReturns()) {
+			source.validateExpression(mas, ids.getVariables(), currConstVars, conditional);
+		}
+		else
+			throw new LanguageException("Unexpected error.");
+		
+		if ( source instanceof FunctionCallIdentifier ) {
+			for (int j =0; j< targetList.size(); j++) {
+				DataIdentifier target = targetList.get(j);
+				// set target properties (based on type info in function call statement return params)
+				FunctionCallIdentifier fci = (FunctionCallIdentifier)source;
+				FunctionStatement fstmt = (FunctionStatement)_dmlProg
+					.getFunctionStatementBlock(fci.getNamespace(), fci.getName()).getStatement(0);
+				if (fstmt == null){
+					fci.raiseValidateError(" function " + fci.getName() 
+						+ " is undefined in namespace " + fci.getNamespace(), conditional);
+				}
+				if (!(target instanceof IndexedIdentifier)){
+					target.setProperties(fstmt.getOutputParams().get(j));
+				}
+				else{
+					DataIdentifier targetAsSeen = ids.getVariable(target.getName());
+					if (targetAsSeen == null){
+						raiseValidateError(target.printErrorLocation() + "cannot assign value to indexed identifier " 
+							+ target.toString() + " without first initializing " + target.getName(), conditional);
+					}
+					target.setProperties(targetAsSeen);
+				}
+				ids.addVariable(target.getName(), target);
+			}
+		}
+		else if ( source instanceof BuiltinFunctionExpression || source instanceof ParameterizedBuiltinFunctionExpression ) {
+			Identifier[] outputs = source.getOutputs();
+			for (int j=0; j < targetList.size(); j++) {
+				ids.addVariable(targetList.get(j).getName(), (DataIdentifier)outputs[j]);
+			}
+		}
+	}
+	
 	public void setStatementFormatType(OutputStatement s, boolean conditionalValidate)
 		throws LanguageException, ParseException
 	{
@@ -944,10 +1101,9 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 		return _constVarsOut;
 	}
 
-
+	@Override
 	public VariableSet analyze(VariableSet loPassed)
-		throws LanguageException{
-
+		throws LanguageException {
 		VariableSet candidateLO = new VariableSet();
 		candidateLO.addVariables(loPassed);
 		//candidateLO.addVariables(_gen);
@@ -1062,9 +1218,10 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 	// materialized hops recompilation / updateinplace flags
 	////
 
-	public void updateRecompilationFlag() throws HopsException {
-		_requiresRecompile = ConfigurationManager.isDynamicRecompilation()
-			                 && Recompiler.requiresRecompilation(get_hops());
+	public boolean updateRecompilationFlag() throws HopsException {
+		return (_requiresRecompile =
+			ConfigurationManager.isDynamicRecompilation()
+			&& Recompiler.requiresRecompilation(getHops()));
 	}
 
 	public boolean requiresRecompilation() {
@@ -1078,5 +1235,4 @@ public class StatementBlock extends LiveVariableAnalysis implements ParseInfo
 	public void setUpdateInPlaceVars( ArrayList<String> vars ) {
 		_updateInPlaceVars = vars;
 	}
-
-}  // end class
+}
