@@ -22,6 +22,7 @@ package org.apache.sysml.runtime.codegen;
 import java.util.Arrays;
 
 import org.apache.commons.math3.util.FastMath;
+import org.apache.sysml.runtime.functionobjects.BitwAnd;
 import org.apache.sysml.runtime.functionobjects.IntegerDivide;
 import org.apache.sysml.runtime.functionobjects.Modulus;
 import org.apache.sysml.runtime.matrix.data.LibMatrixMult;
@@ -37,6 +38,7 @@ public class LibSpoofPrimitives
 {
 	private static IntegerDivide intDiv = IntegerDivide.getFnObject();
 	private static Modulus mod = Modulus.getFnObject();
+	private static BitwAnd bwAnd = BitwAnd.getBitwAndFnObject();
 	
 	//global pool of reusable vectors, individual operations set up their own thread-local
 	//ring buffers of reusable vectors with specific number of vectors and vector sizes 
@@ -118,6 +120,19 @@ public class LibSpoofPrimitives
 		}
 	}
 	
+	public static void vectOuterMultAdd(double[] a, double[] b, double[] c, int ai, int[] bix, int bi, int ci, int blen, int len1, int len2) {
+		if( isFlipOuter(len1, len2) ) {
+			for( int i=bi; i<bi+blen; i++ ) {
+				final int cix = ci + bix[i] * len1;
+				LibMatrixMult.vectMultiplyAdd(b[i], a, c, ai, cix, len1);
+			}
+		}
+		else {
+			for( int i=0, cix=ci; i < len1; i++, cix+=len2 )
+				LibMatrixMult.vectMultiplyAdd(a[ai+i], b, c, bix, bi, cix, blen);
+		}
+	}
+	
 	public static void vectMultAdd(double[] a, double bval, double[] c, int bi, int ci, int len) {
 		if( a == null || bval == 0 ) return;
 		LibMatrixMult.vectMultiplyAdd(bval, a, c, bi, ci, len);
@@ -183,16 +198,68 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectMultWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectMultWrite(b, a, bix, ai, bi, blen, len);
+	}
+	
 	public static void vectWrite(double[] a, double[] c, int ci, int len) {
 		if( a == null ) return;
 		System.arraycopy(a, 0, c, ci, len);
 	}
 	
+	public static void vectWrite(double[] a, double[] c, int ai, int ci, int len) {
+		if( a == null ) return;
+		System.arraycopy(a, ai, c, ci, len);
+	}
+	
 	public static void vectWrite(boolean[] a, boolean[] c, int[] aix) {
 		if( a == null ) return;
 		for( int i=0; i<aix.length; i++ )
 			c[aix[i]] = a[i];
 	}
+	
+	public static void vectWrite(boolean[] a, boolean[] c, int[] aix, int ai, int ci, int alen) {
+		if( a == null ) return;
+		for( int i=ai; i<ai+alen; i++ )
+			c[ci+aix[i]] = a[i];
+	}
+	
+	// cbind handling
+	
+	public static double[] vectCbindAdd(double[] a, double b, double[] c, int ai, int ci, int len) {
+		LibMatrixMult.vectAdd(a, c, ai, ci, len);
+		c[ci+len] += b;
+		return c;
+	}
+	
+	public static double[] vectCbindAdd(double[] a, double b, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		LibMatrixMult.vectAdd(a, c, aix, ai, ci, alen);
+		c[ci+len] += b;
+		return c;
+	}
+	
+	public static double[] vectCbindWrite(double a, double b) {
+		double[] c = allocVector(2, false);
+		c[0] = a;
+		c[1] = b;
+		return c;
+	}
+	
+	public static double[] vectCbindWrite(double[] a, double b, int aix, int len) {
+		double[] c = allocVector(len+1, false);
+		System.arraycopy(a, aix, c, 0, len);
+		c[len] = b;
+		return c;
+	}
+	
+	public static double[] vectCbindWrite(double[] a, double b, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len+1, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = a[j];
+		c[len] = b;
+		return c;
+	}
 
 	// custom vector sums, mins, maxs
 	
@@ -228,8 +295,16 @@ public class LibSpoofPrimitives
 		return vectSum(avals, ai, alen);
 	}
 	
+	public static double vectSumsq(double[] a, int ai, int len) { 
+		return LibMatrixMult.dotProduct(a, a, ai, ai, len);
+	}
+	
+	public static double vectSumsq(double[] avals, int[] aix, int ai, int alen, int len) {
+		return LibMatrixMult.dotProduct(avals, avals, ai, ai, alen);
+	}
+	
 	public static double vectMin(double[] a, int ai, int len) { 
-		double val = Double.MAX_VALUE;
+		double val = Double.POSITIVE_INFINITY;
 		for( int i = ai; i < ai+len; i++ )
 			val = Math.min(a[i], val);
 		return val; 
@@ -241,7 +316,7 @@ public class LibSpoofPrimitives
 	}
 	
 	public static double vectMax(double[] a, int ai, int len) { 
-		double val = -Double.MAX_VALUE;
+		double val = Double.NEGATIVE_INFINITY;
 		for( int i = ai; i < ai+len; i++ )
 			val = Math.max(a[i], val);
 		return val; 
@@ -252,6 +327,18 @@ public class LibSpoofPrimitives
 		return (alen<len) ? Math.max(val, 0) : val;
 	}
 	
+	public static double vectCountnnz(double[] a, int ai, int len) { 
+		int count = 0;
+		for( int i = ai; i < ai+len; i++ )
+			count += (a[i] != 0) ? 1 : 0;
+		return count;
+	} 
+	
+	public static double vectCountnnz(double[] avals, int[] aix, int ai, int alen, int len) {
+		//pure meta data operation
+		return alen;
+	}
+	
 	//custom vector div
 	
 	public static void vectDivAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
@@ -321,6 +408,18 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectDivWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++ ) {
+			double aval = a[bi + j];
+			c[j] = (aval==0) ? Double.NaN : (aval>0) ? 
+				Double.POSITIVE_INFINITY : Double.NEGATIVE_INFINITY;
+		}
+		for( int j = bi; j < bi+blen; j++ )
+			c[bix[j]] = a[ai+bix[j]] / b[j];
+		return c;
+	}
+	
 	//custom vector minus
 	
 	public static void vectMinusAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
@@ -393,6 +492,14 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectMinusWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		double[] c = allocVector(len, false);
+		System.arraycopy(a, ai, c, 0, len);
+		for( int j = bi; j < bi+blen; j++ )
+			c[bix[j]] -= b[j];
+		return c;
+	}
+	
 	//custom vector plus
 	
 	public static void vectPlusAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
@@ -451,8 +558,92 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectPlusWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectPlusWrite(b, a, bix, bi, ai, blen, len);
+	}
+
+	//custom vector xor
+	/**
+	 * Computes c = xor(A,B)
+	 *
+	 * @param a dense input vector A
+	 * @param bval scalar value
+	 * @param c resultant vector
+	 * @param ai start position in A
+	 * @param ci index of c
+	 * @param len number of processed elements
+	 */
+	public static void vectXorAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  ( (a[j] != 0) != (bval != 0) ) ? 1 : 0;
+	}
+
+	public static void vectXorAdd(double bval, double[] a, double[] c, int ai, int ci, int len) {
+		vectXorAdd(a, bval, c, ai, ci, len);
+	}
+
+	public static void vectXorAdd(double[] a, double bval, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += ( (a[j] != 0) != (bval != 0) ) ? 1 : 0;
+	}
+
+	public static void vectXorAdd(double bval, double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		vectXorAdd(a, bval, c, aix, ai, ci, alen, len);
+	}
+
+	//1. scalar vs. dense vector
+	public static double[] vectXorWrite(double[] a, double bval, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++)
+			c[j] = ( ( a[ai+j] != 0) != (bval != 0) ) ? 1 : 0;
+		return c;
+	}
+
+	//2. dense vector vs. scalar
+	public static double[] vectXorWrite(double bval, double[] a, int ai, int len) {
+		return vectXorWrite(a, bval, ai, len);
+	}
+
+	//3. dense vector vs. dense vector
+	public static double[] vectXorWrite(double[] a, double[] b, int ai, int bi, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++)
+			c[j] = ( (a[ai + j] != 0) != (b[bi + j] != 0) ) ? 1 : 0;
+		return c;
+	}
+
+	//4. sparse vector vs scalar
+	public static double[] vectXorWrite(double[] a, double bval, int[] aix, int ai, int alen, int len) {
+		double init = (bval != 0) ? 1 : 0;
+		double[] c = allocVector(len, true, init);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = (a[j] != 0) ? 0 : 1;
+		return c;
+	}
+
+	//5. scalar vs. sparse vector
+	public static double[] vectXorWrite(double bval, double[] a, int[] aix, int ai, int alen, int len) {
+		return vectXorWrite(a, bval, aix, ai, alen, len);
+	}
+
+	//6. sparse vector vs. dense vector
+	public static double[] vectXorWrite(double[] a, double[] b, int[] aix, int ai, int bi, int alen, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++ )
+			c[j] = (b[bi+j] != 0) ? 1 : 0;
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = ( ( a[j] != 0) != (c[aix[j]] != 0) )? 1 : 0;
+		return c;
+	}
+
+	//6. sparse vector vs. dense vector
+	public static double[] vectXorWrite(double[] a, double[] b, int ai, int[] aix, int bi, int alen, int len) {
+		return vectXorWrite(a, b, aix, ai, bi, alen, len);
+	}
+
 	//custom vector pow
-	
+
 	public static void vectPowAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
 		for( int j = ai; j < ai+len; j++, ci++)
 			c[ci] += Math.pow(a[j], bval);
@@ -464,15 +655,16 @@ public class LibSpoofPrimitives
 	}
 
 	public static void vectPowAdd(double[] a, double bval, double[] c, int[] aix, int ai, int ci, int alen, int len) {
-		if( bval == 0 ) //handle 0^0=1
+		if( bval == 0 ) //handle 0^0=1 & a^0=1
 			for( int j=0; j<len; j++ )
 				c[ci + j] += 1;
-		for( int j = ai; j < ai+alen; j++ )
-			c[ci + aix[j]] += Math.pow(a[j], bval) - 1;
+		else //handle 0^b=0 & a^b
+			for( int j = ai; j < ai+alen; j++ )
+				c[ci + aix[j]] += Math.pow(a[j], bval);
 	}
 	
 	public static void vectPowAdd(double bval, double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
-		for( int j=0; j<len; j++ )
+		for( int j=0; j<len; j++ ) //handle 0^0=1 & b^0=1
 			c[ci + j] += 1;
 		for( int j = ai; j < ai+alen; j++ )
 			c[ci + aix[j]] += Math.pow(bval, a[j]) - 1;
@@ -576,6 +768,11 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectMinWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectMinWrite(b, a, bix, bi, ai, blen, len);
+	}
+	
 	//custom vector max
 	
 	public static void vectMaxAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
@@ -637,6 +834,11 @@ public class LibSpoofPrimitives
 			c[aix[j]] = Math.max(a[j], b[bi + aix[j]]);
 		return c;
 	}
+	
+	public static double[] vectMaxWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectMaxWrite(b, a, bix, bi, ai, blen, len);
+	}
 
 	//custom exp
 	
@@ -832,25 +1034,25 @@ public class LibSpoofPrimitives
 	
 	public static void vectLogAdd(double[] a, double[] c, int ai, int ci, int len) {
 		for( int j = ai; j < ai+len; j++, ci++)
-			c[ci] += FastMath.log(a[j]);
+			c[ci] += Math.log(a[j]);
 	}
 
 	public static void vectLogAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
 		for( int j = ai; j < ai+alen; j++ )
-			c[ci + aix[j]] += FastMath.log(a[j]);
+			c[ci + aix[j]] += Math.log(a[j]);
 	}
 	
 	public static double[] vectLogWrite(double[] a, int ai, int len) {
 		double[] c = allocVector(len, false);
 		for( int j = 0; j < len; j++, ai++)
-			c[j] = FastMath.log(a[ai]);
+			c[j] = Math.log(a[ai]);
 		return c;
 	}
 
 	public static double[] vectLogWrite(double[] a, int[] aix, int ai, int alen, int len) {
 		double[] c = allocVector(len, true, Double.NEGATIVE_INFINITY);
 		for( int j = ai; j < ai+alen; j++ )
-			c[aix[j]] = FastMath.log(a[j]);
+			c[aix[j]] = Math.log(a[j]);
 		return c;
 	}
 	
@@ -958,6 +1160,241 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	//custom sin
+	
+	public static void vectSinAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  FastMath.sin(a[j]);
+	}
+
+	public static void vectSinAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += FastMath.sin(a[j]);
+	}
+	
+	public static double[] vectSinWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = FastMath.sin(a[ai]);
+		return c;
+	}
+
+	public static double[] vectSinWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = FastMath.sin(a[j]);
+		return c;
+	}
+	
+	//custom cos
+	
+	public static void vectCosAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  FastMath.cos(a[j]);
+	}
+
+	public static void vectCosAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += FastMath.cos(a[j]);
+	}
+	
+	public static double[] vectCosWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = FastMath.cos(a[ai]);
+		return c;
+	}
+
+	public static double[] vectCosWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true, 1);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = FastMath.cos(a[j]);
+		return c;
+	}
+	
+	//custom tan
+	
+	public static void vectTanAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  FastMath.tan(a[j]);
+	}
+
+	public static void vectTanAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += FastMath.tan(a[j]);
+	}
+	
+	public static double[] vectTanWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = FastMath.tan(a[ai]);
+		return c;
+	}
+
+	public static double[] vectTanWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = FastMath.tan(a[j]);
+		return c;
+	}
+
+	//custom asin
+	
+	public static void vectAsinAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  FastMath.asin(a[j]);
+	}
+
+	public static void vectAsinAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += FastMath.asin(a[j]);
+	}
+	
+	public static double[] vectAsinWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = FastMath.asin(a[ai]);
+		return c;
+	}
+
+	public static double[] vectAsinWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = FastMath.asin(a[j]);
+		return c;
+	}
+	
+	//custom acos
+	
+	public static void vectAcosAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  FastMath.acos(a[j]);
+	}
+
+	public static void vectAcosAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += FastMath.acos(a[j]);
+	}
+	
+	public static double[] vectAcosWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = FastMath.acos(a[ai]);
+		return c;
+	}
+
+	public static double[] vectAcosWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true, Math.PI/2);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = FastMath.acos(a[j]);
+		return c;
+	}
+	
+	//custom atan
+	
+	public static void vectAtanAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  Math.atan(a[j]);
+	}
+
+	public static void vectAtanAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += Math.atan(a[j]);
+	}
+	
+	public static double[] vectAtanWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = Math.atan(a[ai]);
+		return c;
+	}
+
+	public static double[] vectAtanWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = Math.atan(a[j]);
+		return c;
+	}
+
+	
+	//custom sinh
+	
+	public static void vectSinhAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  FastMath.sinh(a[j]);
+	}
+
+	public static void vectSinhAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += FastMath.sinh(a[j]);
+	}
+	
+	public static double[] vectSinhWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = FastMath.sinh(a[ai]);
+		return c;
+	}
+
+	public static double[] vectSinhWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = FastMath.sinh(a[j]);
+		return c;
+	}
+	
+	//custom cosh
+	
+	public static void vectCoshAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  FastMath.cosh(a[j]);
+	}
+
+	public static void vectCoshAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += FastMath.cosh(a[j]);
+	}
+	
+	public static double[] vectCoshWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = FastMath.cosh(a[ai]);
+		return c;
+	}
+
+	public static double[] vectCoshWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true, 1);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = FastMath.cosh(a[j]);
+		return c;
+	}
+	
+	//custom tanh
+	
+	public static void vectTanhAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  FastMath.tanh(a[j]);
+	}
+
+	public static void vectTanhAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += FastMath.tanh(a[j]);
+	}
+	
+	public static double[] vectTanhWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = FastMath.tanh(a[ai]);
+		return c;
+	}
+
+	public static double[] vectTanhWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = FastMath.tanh(a[j]);
+		return c;
+	}
+	
 	//custom sign
 	
 	public static void vectSignAdd(double[] a, double[] c, int ai, int ci, int len) {
@@ -1061,6 +1498,58 @@ public class LibSpoofPrimitives
 			c[aix[j]] = Math.sqrt(a[j]);
 		return c;
 	}
+
+	//custom sprop
+	
+	public static void vectSpropAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] += a[j] * (1 - a[j]);
+	}
+
+	public static void vectSpropAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += a[j] * (1 - a[j]);
+	}
+	
+	public static double[] vectSpropWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = a[j] * (1 - a[j]);
+		return c;
+	}
+
+	public static double[] vectSpropWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = a[j] * (1 - a[j]);
+		return c;
+	}
+	
+	//custom sigmoid
+	
+	public static void vectSigmoidAdd(double[] a, double[] c, int ai, int ci, int len) {
+		for( int j = ai; j < ai+len; j++, ci++)
+			c[ci] +=  1 / (1 + FastMath.exp(-a[j]));
+	}
+
+	public static void vectSigmoidAdd(double[] a, double[] c, int[] aix, int ai, int ci, int alen, int len) {
+		for( int j = ai; j < ai+alen; j++ )
+			c[ci + aix[j]] += 1 / (1 + FastMath.exp(-a[j]));
+	}
+	
+	public static double[] vectSigmoidWrite(double[] a, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++, ai++)
+			c[j] = 1 / (1 + FastMath.exp(-a[j]));
+		return c;
+	}
+
+	public static double[] vectSigmoidWrite(double[] a, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true, 0.5); //sigmoid(0) = 0.5
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = 1 / (1 + FastMath.exp(-a[j]));
+		return c;
+	}
 	
 	//custom vector equal
 	
@@ -1126,6 +1615,11 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectEqualWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectEqualWrite(b, a, bix, bi, ai, blen, len);
+	}
+	
 	//custom vector not equal
 	
 	public static void vectNotequalAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
@@ -1189,6 +1683,11 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectNotequalWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectNotequalWrite(b, a, bix, bi, ai, blen, len);
+	}
+	
 	//custom vector less
 	
 	public static void vectLessAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
@@ -1252,6 +1751,11 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectLessWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectGreaterequalWrite(b, a, bix, bi, ai, blen, len);
+	}
+	
 	//custom vector less equal
 	
 	public static void vectLessequalAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
@@ -1314,6 +1818,11 @@ public class LibSpoofPrimitives
 			c[aix[j]] = (a[j] <= b[bi+aix[j]]) ? 1 : 0;
 		return c;
 	}
+	
+	public static double[] vectLessequalWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectGreaterWrite(b, a, bix, bi, ai, blen, len);
+	}
 
 	//custom vector greater
 	
@@ -1378,6 +1887,11 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectGreaterWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectLessequalWrite(b, a, bix, bi, ai, blen, len);
+	}
+	
 	//custom vector greater equal
 	
 	public static void vectGreaterequalAdd(double[] a, double bval, double[] c, int ai, int ci, int len) {
@@ -1441,6 +1955,61 @@ public class LibSpoofPrimitives
 		return c;
 	}
 	
+	public static double[] vectGreaterequalWrite(double[] a, double[] b, int ai, int[] bix, int bi, int blen, int len) {
+		//invariant to the ordering of inputs
+		return vectLessWrite(b, a, bix, bi, ai, blen, len);
+	}
+
+	//bitwise and
+	
+	//1. dense vector vs. scalar
+	public static double[] vectBitwandWrite(double[] a, double bval, int ai, int len) {
+		double[] c = allocVector(len, false);
+		for( int j = 0; j < len; j++ )
+			c[j] = bwAnd(a[ai+j], bval);
+		return c;
+	}
+
+	//2. scalar vs. dense vector
+	public static double[] vectBitwandWrite(double bval, double[] a, int ai, int len) {
+		return vectBitwandWrite(a, bval, ai, len);
+	}
+
+	//3. dense vector vs. dense vector
+	public static double[] vectBitwandWrite(double[] a, double[] b, int ai, int bi, int len) {
+		double[] c= allocVector(len, false);
+		for( int j = 0; j < len; j++ )
+			c[j] = bwAnd(a[ai+j], b[bi+j]);
+		return c;
+	}
+
+	//4. sparse vector vs. scalar.
+	public static double[] vectBitwandWrite(double[] a, double bval, int[] aix, int ai, int alen, int len) {
+		double[] c = allocVector(len, true);
+		int bval1 = (int)bval;
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = bwAnd(a[j], bval1);
+		return c;
+	}
+
+	//5. scalar vs. sparse vector
+	public static double[] vectBitwandWrite(double bval, double[] a, int[] aix, int ai, int alen, int len) {
+		return vectBitwandWrite(a, bval, aix, ai, alen, len);
+	}
+
+	//6. sparse vector vs. dense vector
+	public static double[] vectBitwandWrite(double[] a, double[] b, int[] aix, int ai, int bi, int alen, int len) {
+		double[] c = allocVector(len, true);
+		for( int j = ai; j < ai+alen; j++ )
+			c[aix[j]] = bwAnd(a[j], b[bi+aix[j]]);
+		return c;
+	}
+
+	//6. sparse vector vs. dense vector
+	public static double[] vectBitwandWrite(double[] a, double[] b, int ai, int[] aix, int bi, int alen, int len) {
+		return vectBitwandWrite(a, b, aix, ai, bi, alen, len);
+	}
+
 	//complex builtin functions that are not directly generated
 	//(included here in order to reduce the number of imports)
 	
@@ -1452,6 +2021,10 @@ public class LibSpoofPrimitives
 		return mod.execute(in1, in2);
 	}
 	
+	public static double bwAnd(double in1, double in2) {
+		return bwAnd.execute(in1, in2);
+	}
+	
 	public static boolean isFlipOuter(int len1, int len2) {
 		return (len1 > 64 * len2);
 	}
@@ -1459,23 +2032,25 @@ public class LibSpoofPrimitives
 	//dynamic memory management
 	
 	public static void setupThreadLocalMemory(int numVectors, int len) {
-		setupThreadLocalMemory(numVectors, len, -1);
+		if( numVectors > 0 )
+			setupThreadLocalMemory(numVectors, len, -1);
 	}
 	
 	public static void setupThreadLocalMemory(int numVectors, int len, int len2) {
-		memPool.set(new VectorBuffer(numVectors, len, len2));
+		if( numVectors > 0 )
+			memPool.set(new VectorBuffer(numVectors, len, len2));
 	}
 	
 	public static void cleanupThreadLocalMemory() {
 		memPool.remove();
 	}
 	
-	protected static double[] allocVector(int len, boolean reset) {
+	public static double[] allocVector(int len, boolean reset) {
 		return allocVector(len, reset, 0);
 	}
 	
 	protected static double[] allocVector(int len, boolean reset, double resetVal) {
-		VectorBuffer buff = memPool.get(); 
+		VectorBuffer buff = memPool.get();
 		
 		//find next matching vector in ring buffer or
 		//allocate new vector if required
@@ -1494,15 +2069,21 @@ public class LibSpoofPrimitives
 	 * vectors of different sizes are interspersed.
 	 */
 	private static class VectorBuffer {
+		private static final int MAX_SIZE = 512*1024; //4MB
 		private final double[][] _data;
 		private int _pos;
 		private int _len1;
 		private int _len2;
 		
 		public VectorBuffer(int num, int len1, int len2) {
-			int lnum = (len2 > 0) ? 2*num : num;
+			//best effort size restriction since large intermediates
+			//not necessarily used (num refers to the total number)
+			len1 = Math.min(len1, MAX_SIZE);
+			len2 = Math.min(len2, MAX_SIZE);
+			//pre-allocate ring buffer
+			int lnum = (len2>0 && len1!=len2) ? 2*num : num;
 			_data = new double[lnum][];
-			for( int i=0; i<num; i++ )
+			for( int i=0; i<num; i++ ) {
 				if( lnum > num ) {
 					_data[2*i] = new double[len1];
 					_data[2*i+1] = new double[len2];
@@ -1510,6 +2091,7 @@ public class LibSpoofPrimitives
 				else {
 					_data[i] = new double[len1];
 				}
+			}
 			_pos = -1;
 			_len1 = len1;
 			_len2 = len2;
@@ -1522,5 +2104,11 @@ public class LibSpoofPrimitives
 			} while( _data[_pos].length!=len );
 			return _data[_pos];
 		}
+		@SuppressWarnings("unused")
+		public boolean isReusable(int num, int len1, int len2) {
+			int lnum = (len2>0 && len1!=len2) ? 2*num : num;
+			return (_len1 == len1 && _len2 == len2
+				&& _data.length == lnum);
+		}
 	}
 }
