@@ -24,7 +24,6 @@ import java.util.Iterator;
 import java.util.List;
 import java.util.concurrent.Callable;
 import java.util.concurrent.ExecutorService;
-import java.util.concurrent.Executors;
 import java.util.concurrent.Future;
 import java.util.stream.IntStream;
 
@@ -33,12 +32,15 @@ import org.apache.sysml.runtime.compress.BitmapEncoder;
 import org.apache.sysml.runtime.compress.CompressedMatrixBlock;
 import org.apache.sysml.runtime.instructions.cp.DoubleObject;
 import org.apache.sysml.runtime.instructions.cp.ScalarObject;
+import org.apache.sysml.runtime.matrix.data.DenseBlock;
+import org.apache.sysml.runtime.matrix.data.DenseBlockFactory;
 import org.apache.sysml.runtime.matrix.data.LibMatrixMult;
 import org.apache.sysml.runtime.matrix.data.LibMatrixReorg;
 import org.apache.sysml.runtime.matrix.data.MatrixBlock;
 import org.apache.sysml.runtime.matrix.data.SparseBlock;
 import org.apache.sysml.runtime.matrix.data.SparseRow;
 import org.apache.sysml.runtime.matrix.data.SparseRowVector;
+import org.apache.sysml.runtime.util.CommonThreadPool;
 
 
 public abstract class SpoofRowwise extends SpoofOperator
@@ -72,7 +74,7 @@ public abstract class SpoofRowwise extends SpoofOperator
 		}
 		public boolean isConstDim2(long dim2) {
 			return (this == NO_AGG_CONST || this == COL_AGG_CONST)
-				|| (dim2>0 && isRowTypeB1());
+				|| (dim2>=0 && isRowTypeB1());
 		}
 	}
 	
@@ -137,7 +139,7 @@ public abstract class SpoofRowwise extends SpoofOperator
 			getMinColsMatrixSideInputs(inputs) : -1;
 		if( !aggIncr || !out.isAllocated() )
 			allocateOutputMatrix(m, n, n2, out);
-		double[] c = out.getDenseBlock();
+		DenseBlock c = out.getDenseBlock();
 		final boolean flipOut = _type.isRowTypeB1ColumnAgg()
 			&& LibSpoofPrimitives.isFlipOuter(out.getNumRows(), out.getNumColumns());
 		
@@ -161,13 +163,15 @@ public abstract class SpoofRowwise extends SpoofOperator
 		//post-processing
 		if( allocTmp &&_reqVectMem > 0 )
 			LibSpoofPrimitives.cleanupThreadLocalMemory();
-		out.recomputeNonZeros();
 		if( flipOut ) {
 			fixTransposeDimensions(out);
 			out = LibMatrixReorg.transpose(out, new MatrixBlock(
 				out.getNumColumns(), out.getNumRows(), false));
 		}
-		out.examSparsity();
+		if( !aggIncr ) {
+			out.recomputeNonZeros();
+			out.examSparsity();
+		}
 		return out;
 	}
 	
@@ -201,10 +205,10 @@ public abstract class SpoofRowwise extends SpoofOperator
 		double[] scalars = prepInputScalars(scalarObjects);
 		
 		//core parallel execute
-		ExecutorService pool = Executors.newFixedThreadPool( k );
+		ExecutorService pool = CommonThreadPool.get(k);
 		ArrayList<Integer> blklens = (a instanceof CompressedMatrixBlock) ?
 			LibMatrixMult.getAlignedBlockSizes(m, k, BitmapEncoder.BITMAP_BLOCK_SZ) :
-			LibMatrixMult.getBalancedBlockSizesDefault(m, k, false);
+			LibMatrixMult.getBalancedBlockSizesDefault(m, k, (long)m*n<16*PAR_NUMCELL_THRESHOLD);
 		
 		try
 		{
@@ -214,11 +218,11 @@ public abstract class SpoofRowwise extends SpoofOperator
 				int outLen = out.getNumRows() * out.getNumColumns();
 				for( int i=0, lb=0; i<blklens.size(); lb+=blklens.get(i), i++ )
 					tasks.add(new ParColAggTask(a, b, scalars, n, n2, outLen, lb, lb+blklens.get(i)));
-				List<Future<double[]>> taskret = pool.invokeAll(tasks);
+				List<Future<DenseBlock>> taskret = pool.invokeAll(tasks);
 				//aggregate partial results
 				int len = _type.isColumnAgg() ? out.getNumRows()*out.getNumColumns() : 1;
-				for( Future<double[]> task : taskret )
-					LibMatrixMult.vectAdd(task.get(), out.getDenseBlock(), 0, 0, len);
+				for( Future<DenseBlock> task : taskret )
+					LibMatrixMult.vectAdd(task.get().valuesAt(0), out.getDenseBlockValues(), 0, 0, len);
 				out.recomputeNonZeros();
 			}
 			else {
@@ -284,65 +288,66 @@ public abstract class SpoofRowwise extends SpoofOperator
 		int rlen = out.getNumRows();
 		out.setNumRows(out.getNumColumns());
 		out.setNumColumns(rlen);
+		out.setNonZeros(out.getNumRows()*out.getNumColumns());
 	}
 	
-	private void executeDense(double[] a, SideInput[] b, double[] scalars, double[] c, int n, int rl, int ru) 
-	{
-		if( a == null )
+	private void executeDense(DenseBlock a, SideInput[] b, double[] scalars, DenseBlock c, int n, int rl, int ru) {
+		//forward empty block to sparse
+		if( a == null ) {
+			executeSparse(null, b, scalars, c, n, rl, ru);
 			return;
+		}
 		
 		SideInput[] lb = createSparseSideInputs(b, true);
-		for( int i=rl, aix=rl*n; i<ru; i++, aix+=n ) {
-			//call generated method
-			genexec( a, aix, lb, scalars, c, n, i );
+		for( int i=rl; i<ru; i++ ) {
+			genexec(a.values(i), a.pos(i), lb, scalars,
+				c.values(i), c.pos(i), n, i );
 		}
 	}
 	
-	private void executeSparse(SparseBlock sblock, SideInput[] b, double[] scalars, double[] c, int n, int rl, int ru) 
-	{
+	private void executeSparse(SparseBlock a, SideInput[] b, double[] scalars, DenseBlock c, int n, int rl, int ru) {
 		SideInput[] lb = createSparseSideInputs(b, true);
 		SparseRow empty = new SparseRowVector(1);
 		for( int i=rl; i<ru; i++ ) {
-			if( sblock!=null && !sblock.isEmpty(i) ) {
-				double[] avals = sblock.values(i);
-				int[] aix = sblock.indexes(i);
-				int apos = sblock.pos(i);
-				int alen = sblock.size(i);
-				
+			if( a!=null && !a.isEmpty(i) ) {
 				//call generated method
-				genexec(avals, aix, apos, lb, scalars, c, alen, n, i);
+				genexec(a.values(i), a.indexes(i), a.pos(i), lb, scalars,
+					c.values(i), c.pos(i), a.size(i), n, i);
 			}
 			else
-				genexec(empty.values(),
-					empty.indexes(), 0, lb, scalars, c, 0, n, i);
+				genexec(empty.values(), empty.indexes(), 0, lb, scalars,
+					c.values(i), c.pos(i), 0, n, i);
 		}
 	}
 	
-	private void executeCompressed(CompressedMatrixBlock a, SideInput[] b, double[] scalars, double[] c, int n, int rl, int ru) 
-	{
-		if( a.isEmptyBlock(false) )
+	private void executeCompressed(CompressedMatrixBlock a, SideInput[] b, double[] scalars, DenseBlock c, int n, int rl, int ru) {
+		//forward empty block to sparse
+		if( a.isEmptyBlock(false) ) {
+			executeSparse(null, b, scalars, c, n, rl, ru);
 			return;
+		}
 		
 		SideInput[] lb = createSparseSideInputs(b, true);
 		Iterator<double[]> iter = a.getDenseRowIterator(rl, ru);
 		for( int i=rl; iter.hasNext(); i++ ) {
-			genexec(iter.next(), 0, lb, scalars, c, n, i);
+			genexec(iter.next(), 0, lb, scalars,
+				c.values(i), c.pos(i), n, i);
 		}
 	}
 	
 	//methods to be implemented by generated operators of type SpoofRowAggrgate 
 	
 	protected abstract void genexec(double[] a, int ai, 
-		SideInput[] b, double[] scalars, double[] c, int len, int rowIndex);
+		SideInput[] b, double[] scalars, double[] c, int ci, int len, int rowIndex);
 	
 	protected abstract void genexec(double[] avals, int[] aix, int ai, 
-		SideInput[] b, double[] scalars, double[] c, int alen, int n, int rowIndex);
+		SideInput[] b, double[] scalars, double[] c, int ci, int alen, int n, int rowIndex);
 
 	
 	/**
 	 * Task for multi-threaded column aggregation operations.
 	 */
-	private class ParColAggTask implements Callable<double[]> 
+	private class ParColAggTask implements Callable<DenseBlock> 
 	{
 		private final MatrixBlock _a;
 		private final SideInput[] _b;
@@ -362,12 +367,12 @@ public abstract class SpoofRowwise extends SpoofOperator
 		}
 		
 		@Override
-		public double[] call() throws DMLRuntimeException {
+		public DenseBlock call() throws DMLRuntimeException {
 			
 			//allocate vector intermediates and partial output
 			if( _reqVectMem > 0 )
 				LibSpoofPrimitives.setupThreadLocalMemory(_reqVectMem, _clen, _clen2);
-			double[] c = new double[_outLen];
+			DenseBlock c = DenseBlockFactory.createDenseBlock(1, _outLen);
 			
 			if( _a instanceof CompressedMatrixBlock )
 				executeCompressed((CompressedMatrixBlock)_a, _b, _scalars, c, _clen, _rl, _ru);
