@@ -38,6 +38,7 @@ import org.apache.sysml.parser.Expression.DataType;
 import org.apache.sysml.parser.Expression.ValueType;
 import org.apache.sysml.runtime.DMLRuntimeException;
 import org.apache.sysml.runtime.controlprogram.caching.LazyWriteBuffer.RPolicy;
+import org.apache.sysml.runtime.controlprogram.parfor.stat.InfrastructureAnalyzer;
 import org.apache.sysml.runtime.controlprogram.parfor.util.IDSequence;
 import org.apache.sysml.runtime.instructions.cp.CPInstruction;
 import org.apache.sysml.runtime.instructions.cp.Data;
@@ -47,13 +48,12 @@ import org.apache.sysml.runtime.instructions.spark.data.BroadcastObject;
 import org.apache.sysml.runtime.instructions.spark.data.RDDObject;
 import org.apache.sysml.runtime.io.IOUtilFunctions;
 import org.apache.sysml.runtime.matrix.MatrixCharacteristics;
-import org.apache.sysml.runtime.matrix.MatrixDimensionsMetaData;
-import org.apache.sysml.runtime.matrix.MatrixFormatMetaData;
+import org.apache.sysml.runtime.matrix.MetaDataFormat;
+import org.apache.sysml.runtime.matrix.MetaDataNumItemsByEachReducer;
 import org.apache.sysml.runtime.matrix.MetaData;
 import org.apache.sysml.runtime.matrix.data.FileFormatProperties;
 import org.apache.sysml.runtime.matrix.data.InputInfo;
 import org.apache.sysml.runtime.matrix.data.MatrixBlock;
-import org.apache.sysml.runtime.matrix.data.NumItemsByEachReducerMetaData;
 import org.apache.sysml.runtime.matrix.data.OutputInfo;
 import org.apache.sysml.runtime.util.LocalFileUtils;
 import org.apache.sysml.runtime.util.MapReduceTool;
@@ -108,40 +108,40 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
     	MODIFY, 
     	CACHED,
     	CACHED_NOWRITE,
-    };
+    }
 	
 	/** Global flag indicating if caching is enabled (controls eviction) */
-	private static boolean _activeFlag = false;
+	private static volatile boolean _activeFlag = false;
 	
 	/** Global sequence for generating unique ids. */
 	private static IDSequence _seq = null;   
 
 	// Global eviction path and prefix (prefix used for isolation purposes)
-    public static String cacheEvictionLocalFilePath = null; //set during init
-    public static String cacheEvictionLocalFilePrefix = "cache";
+	public static String cacheEvictionLocalFilePath = null; //set during init
+	public static String cacheEvictionLocalFilePrefix = "cache";
 
 	/**
 	 * Current state of pinned variables, required for guarded collect.
 	 */
 	private static ThreadLocal<Long> sizePinned = new ThreadLocal<Long>() {
-        @Override protected Long initialValue() { return 0L; }
-    };
+		@Override protected Long initialValue() { return 0L; }
+	};
 
 	//current size of live broadcast objects (because Spark's ContextCleaner maintains 
 	//a buffer with references to prevent eager cleanup by GC); note that this is an 
 	//overestimate, because we maintain partitioned broadcasts as soft references, which 
 	//might be collected by the GC and subsequently cleaned up by Spark's ContextCleaner.
-	private static AtomicLong _refBCs = new AtomicLong(0);	
-    
+	private static final AtomicLong _refBCs = new AtomicLong(0);
+
 	static {
 		_seq = new IDSequence();
 	}
-		
+
 	/**
 	 * The unique (JVM-wide) ID of a cacheable data object; to ensure unique IDs across JVMs, we
 	 * concatenate filenames with a unique prefix (map task ID). 
 	 */
-	private final int _uniqueID;
+	private final long _uniqueID;
 	
 	/** The cache status of the data blob (whether it can be or is evicted, etc. */
 	private CacheStatus _cacheStatus = null;
@@ -187,7 +187,6 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	// additional private flags and meta data
 	private int     _numReadThreads = 0;   //number of threads for read from HDFS
 	private boolean _cleanupFlag = true;   //flag if obj unpinned (cleanup enabled)	
-	private String  _varName = "";         //plan variable name
 	private String  _cacheFileName = null; //local eviction file name
 	private boolean _requiresLocalWrite = false; //flag if local write for read obj
 	private boolean _isAcquireFromEmpty = false; //flag if read from status empty 
@@ -206,11 +205,11 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	 * @param vt value type
 	 */
 	protected CacheableData(DataType dt, ValueType vt) {
-		super (dt, vt);		
-		_uniqueID = (int)_seq.getNextID();		
+		super (dt, vt);
+		_uniqueID = isCachingActive() ? _seq.getNextID() : -1;
 		_cacheStatus = CacheStatus.EMPTY;
 		_numReadThreads = 0;
-		_gpuObjects = new HashMap<GPUContext, GPUObject>();
+		_gpuObjects = new HashMap<>();
 	}
 	
 	/**
@@ -223,7 +222,6 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		_cleanupFlag = that._cleanupFlag;
 		_hdfsFileName = that._hdfsFileName;
 		_hdfsFileExists = that._hdfsFileExists; 
-		_varName = that._varName;
 		_gpuObjects = that._gpuObjects;
 	}
 
@@ -248,14 +246,6 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		return _cleanupFlag;
 	}
 
-	public void setVarName(String s) {
-		_varName = s;
-	}
-
-	public String getVarName() {
-		return _varName;
-	}
-
 	public boolean isHDFSFileExists() {
 		return _hdfsFileExists;
 	}
@@ -314,8 +304,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	}
 	
 	public MatrixCharacteristics getMatrixCharacteristics() {
-		MatrixDimensionsMetaData meta = (MatrixDimensionsMetaData) _metaData;
-		return meta.getMatrixCharacteristics();
+		return _metaData.getMatrixCharacteristics();
 	}
 
 	public abstract void refreshMetaData() 
@@ -386,7 +375,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		throws CacheException
 	{
 		if( LOG.isTraceEnabled() )
-			LOG.trace("Acquire read "+getVarName());
+			LOG.trace("Acquire read "+hashCode());
 		long t0 = DMLScript.STATISTICS ? System.nanoTime() : 0;
 		
 		if ( !isAvailableToRead() )
@@ -404,7 +393,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
                 LOG.error("Inconsistent internal state - A copy of this CacheableData was dirty on more than 1 GPU");
                 throw new CacheException("Internal Error : Inconsistent internal state, A copy of this CacheableData was dirty on more than 1 GPU");
             } else if (gObj != null){
-                copiedFromGPU = gObj.acquireHostRead();
+                copiedFromGPU = gObj.acquireHostRead(null);
                 if( _data == null )
                     getCache();
             }
@@ -445,7 +434,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 				setDirty(false);
 			}
 			catch (IOException e) {
-				throw new CacheException("Reading of " + _hdfsFileName + " ("+getVarName()+") failed.", e);
+				throw new CacheException("Reading of " + _hdfsFileName + " ("+hashCode()+") failed.", e);
 			}
 			
 			_isAcquireFromEmpty = true;
@@ -483,7 +472,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		throws CacheException
 	{
 		if( LOG.isTraceEnabled() )
-			LOG.trace("Acquire modify "+getVarName());
+			LOG.trace("Acquire modify "+hashCode());
 		long t0 = DMLScript.STATISTICS ? System.nanoTime() : 0;
 		
 		if ( !isAvailableToModify() )
@@ -507,7 +496,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 			}
 			catch (IOException e)
 			{
-				throw new CacheException("Reading of " + _hdfsFileName + " ("+getVarName()+") failed.", e);
+				throw new CacheException("Reading of " + _hdfsFileName + " ("+hashCode()+") failed.", e);
 			}
 		}
 
@@ -546,7 +535,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		throws DMLRuntimeException
 	{
 		if( LOG.isTraceEnabled() )
-			LOG.trace("Acquire modify newdata "+getVarName());
+			LOG.trace("Acquire modify newdata "+hashCode());
 		long t0 = DMLScript.STATISTICS ? System.nanoTime() : 0;
 		
 		if (! isAvailableToModify ())
@@ -604,7 +593,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		throws CacheException
 	{
 		if( LOG.isTraceEnabled() )
-			LOG.trace("Release "+getVarName());
+			LOG.trace("Release "+hashCode());
 		long t0 = DMLScript.STATISTICS ? System.nanoTime() : 0;
 		
 		boolean write = false;
@@ -644,7 +633,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 					}
 				}
 				catch (Exception e) {
-					throw new CacheException("Eviction to local path " + filePath + " ("+getVarName()+") failed.", e);
+					throw new CacheException("Eviction to local path " + filePath + " ("+hashCode()+") failed.", e);
 				}
 				_requiresLocalWrite = false;
 			}
@@ -654,7 +643,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 			_data = null;			
 		}
 		else if( LOG.isTraceEnabled() ){
-			LOG.trace("Var "+getVarName()+" not subject to caching, state="+getStatusAsString());
+			LOG.trace("Var "+hashCode()+" not subject to caching, state="+getStatusAsString());
 		}
 
 		if( DMLScript.STATISTICS ){
@@ -663,8 +652,6 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		}
 	}
 	
-	protected void clearReusableData() {}
-	
 	/**
 	 * Sets the cache block reference to <code>null</code>, abandons the old block.
 	 * Makes the "envelope" empty.  Run it to finalize the object (otherwise the
@@ -679,7 +666,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		throws DMLRuntimeException
 	{
 		if( LOG.isTraceEnabled() )
-			LOG.trace("Clear data "+getVarName());
+			LOG.trace("Clear data "+hashCode());
 		
 		// check if cleanup enabled and possible 
 		if( !isCleanupEnabled() ) 
@@ -694,8 +681,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 			freeEvictedBlob();	
 		
 		// clear the in-memory data
-		clearReusableData();
-		_data = null;	
+		_data = null;
 		clearCache();
 		
 		// clear rdd/broadcast back refs
@@ -703,13 +689,10 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 			_rddHandle.setBackReference(null);
 		if( _bcHandle != null )
 			_bcHandle.setBackReference(null);
-		if( _gpuObjects != null ) {
-		    for (GPUObject gObj : _gpuObjects.values()){
-		        if (gObj != null) {
-                    gObj.clearData();
-                }
-            }
-        }
+		if( _gpuObjects != null )
+			for (GPUObject gObj : _gpuObjects.values())
+				if (gObj != null)
+					gObj.clearData();
 
 		// change object state EMPTY
 		setDirty(false);
@@ -776,7 +759,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		throws CacheException
 	{
 		if( LOG.isTraceEnabled() )
-			LOG.trace("Export data "+getVarName()+" "+fName);
+			LOG.trace("Export data "+hashCode()+" "+fName);
 		long t0 = DMLScript.STATISTICS ? System.nanoTime() : 0;
 		
 		//prevent concurrent modifications
@@ -793,7 +776,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
                 LOG.error("Inconsistent internal state - A copy of this CacheableData was dirty on more than 1 GPU");
                 throw new CacheException("Internal Error : Inconsistent internal state, A copy of this CacheableData was dirty on more than 1 GPU");
             } else if (gObj != null){
-                copiedFromGPU = gObj.acquireHostRead();
+                copiedFromGPU = gObj.acquireHostRead(null);
                 if( _data == null )
                     getCache();
             }
@@ -828,7 +811,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 				}
 				catch (IOException e)
 				{
-				    throw new CacheException("Reading of " + _hdfsFileName + " ("+getVarName()+") failed.", e);
+				    throw new CacheException("Reading of " + _hdfsFileName + " ("+hashCode()+") failed.", e);
 				}
 			}
 			//get object from cache
@@ -928,7 +911,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		long begin = LOG.isTraceEnabled() ? System.currentTimeMillis() : 0;
 		
 		if( LOG.isTraceEnabled() )
-			LOG.trace ("CACHE: Restoring matrix...  " + getVarName() + "  HDFS path: " + 
+			LOG.trace ("CACHE: Restoring matrix...  " + hashCode() + "  HDFS path: " + 
 						(_hdfsFileName == null ? "null" : _hdfsFileName) + ", Restore from path: " + cacheFilePathAndName);
 				
 		if (_data != null)
@@ -961,7 +944,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		String cacheFilePathAndName = getCacheFilePathAndName();
 		long begin = LOG.isTraceEnabled() ? System.currentTimeMillis() : 0;
 		if( LOG.isTraceEnabled() )
-			LOG.trace("CACHE: Freeing evicted matrix...  " + getVarName() + "  HDFS path: " + 
+			LOG.trace("CACHE: Freeing evicted matrix...  " + hashCode() + "  HDFS path: " + 
 						(_hdfsFileName == null ? "null" : _hdfsFileName) + " Eviction path: " + cacheFilePathAndName);
 		
 		LazyWriteBuffer.deleteBlock(cacheFilePathAndName);
@@ -974,6 +957,10 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		return (_data.getInMemorySize() <= CACHING_THRESHOLD);
 	}
 	
+	public long getDataSize() {
+		return (_data != null) ?_data.getInMemorySize() : 0;
+	}
+	
 	protected ValueType[] getSchema() {
 		return null;
 	}
@@ -984,13 +971,13 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		String debugNameEnding = (_hdfsFileName == null ? "null" : 
 			(_hdfsFileName.length() < maxLength ? _hdfsFileName : "..." + 
 				_hdfsFileName.substring (_hdfsFileName.length() - maxLength + 3)));
-		return getVarName() + " " + debugNameEnding;
+		return hashCode() + " " + debugNameEnding;
 	}
 
 	protected T readBlobFromHDFS(String fname) 
 		throws IOException 
 	{
-		MatrixFormatMetaData iimd = (MatrixFormatMetaData) _metaData;
+		MetaDataFormat iimd = (MetaDataFormat) _metaData;
 		MatrixCharacteristics mc = iimd.getMatrixCharacteristics();
 		return readBlobFromHDFS(fname, mc.getRows(), mc.getCols());
 	}
@@ -1010,7 +997,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	protected void writeMetaData (String filePathAndName, String outputFormat, FileFormatProperties formatProperties)
 		throws DMLRuntimeException, IOException
 	{		
-		MatrixFormatMetaData iimd = (MatrixFormatMetaData) _metaData;
+		MetaDataFormat iimd = (MetaDataFormat) _metaData;
 	
 		if (iimd == null)
 			throw new DMLRuntimeException("Unexpected error while writing mtd file (" + filePathAndName + ") -- metadata is null.");
@@ -1042,7 +1029,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		boolean ret = true;
 		if( outputFormat != null ) {
 			try {
-				MatrixFormatMetaData iimd = (MatrixFormatMetaData) _metaData;
+				MetaDataFormat iimd = (MetaDataFormat) _metaData;
 				OutputInfo oi1 = InputInfo.getMatchingOutputInfo( iimd.getInputInfo() );
 				OutputInfo oi2 = OutputInfo.stringToOutputInfo( outputFormat );
 				if( oi1 != oi2 )
@@ -1059,17 +1046,13 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	
 	// ------------- IMPLEMENTED CACHE LOGIC METHODS --------------	
 	
-	protected int getUniqueCacheID() {
-		return _uniqueID;
-	}
-
 	protected String getCacheFilePathAndName () {
 		if( _cacheFileName==null ) {
 			StringBuilder sb = new StringBuilder();
 			sb.append(CacheableData.cacheEvictionLocalFilePath); 
 			sb.append(CacheableData.cacheEvictionLocalFilePrefix);
-			sb.append(String.format ("%09d", getUniqueCacheID()));
-			sb.append(CacheableData.CACHING_EVICTION_FILEEXTENSION);			
+			sb.append(String.format ("%09d", _uniqueID));
+			sb.append(CacheableData.CACHING_EVICTION_FILEEXTENSION);
 			_cacheFileName = sb.toString();
 		}
 		
@@ -1254,7 +1237,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	 * referenced cache block.  
 	 */
 	protected void createCache( ) {
-		_cache = new SoftReference<T>( _data );	
+		_cache = new SoftReference<>( _data );	
 	}
 
 	/**
@@ -1284,7 +1267,7 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		sizePinned.set( Math.max(size,0) );
 	}
 
-	protected long getPinnedSize() {
+	protected static long getPinnedSize() {
 		return sizePinned.get();
 	}
 	
@@ -1293,7 +1276,10 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 	}
 	
 	public static long getBroadcastSize() {
-		return _refBCs.longValue();
+		//scale the total sum of all broadcasts by the current fraction
+		//of local memory to equally distribute it across parfor workers
+		return (long) (_refBCs.longValue() *
+			InfrastructureAnalyzer.getLocalMaxMemoryFraction());
 	}
 	
 	// --------- STATIC CACHE INIT/CLEANUP OPERATIONS ----------
@@ -1373,15 +1359,15 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		_activeFlag = true; //turn on caching
 	}
 	
-	public static synchronized boolean isCachingActive() {
+	public static boolean isCachingActive() {
 		return _activeFlag;
 	}
 	
-	public static synchronized void disableCaching() {
+	public static void disableCaching() {
 		_activeFlag = false;
 	}
 	
-	public static synchronized void enableCaching() {
+	public static void enableCaching() {
 		_activeFlag = true;
 	}
 
@@ -1420,19 +1406,20 @@ public abstract class CacheableData<T extends CacheBlock> extends Data
 		return ret;
 	}
 
+	@Override
 	public String toString() {
 		StringBuilder str = new StringBuilder();
 		str.append(getClass().getSimpleName());
 		str.append(": ");
 		str.append(_hdfsFileName + ", ");
 
-		if (_metaData instanceof NumItemsByEachReducerMetaData) {
+		if (_metaData instanceof MetaDataNumItemsByEachReducer) {
 			str.append("NumItemsByEachReducerMetaData");
 		} else {
 			try {
-				MatrixFormatMetaData md = (MatrixFormatMetaData) _metaData;
+				MetaDataFormat md = (MetaDataFormat) _metaData;
 				if (md != null) {
-					MatrixCharacteristics mc = ((MatrixDimensionsMetaData) _metaData).getMatrixCharacteristics();
+					MatrixCharacteristics mc = _metaData.getMatrixCharacteristics();
 					str.append(mc.toString());
 
 					InputInfo ii = md.getInputInfo();
