@@ -37,6 +37,8 @@ import scala.reflect.ClassManifestFactory;
 import scala.runtime.AbstractFunction1;
 
 import org.apache.sysml.hops.AggBinaryOp.SparkAggType;
+import org.apache.sysml.lops.LeftIndex;
+import org.apache.sysml.lops.RightIndex;
 import org.apache.sysml.lops.LeftIndex.LixCacheType;
 import org.apache.sysml.hops.OptimizerUtils;
 import org.apache.sysml.runtime.DMLRuntimeException;
@@ -59,17 +61,7 @@ import org.apache.sysml.runtime.util.IndexRange;
 import org.apache.sysml.runtime.util.UtilFunctions;
 
 /**
- * This class implements the matrix indexing functionality inside CP.  
- * Example instructions: 
- *     rangeReIndex:mVar1:Var2:Var3:Var4:Var5:mVar6
- *         input=mVar1, output=mVar6, 
- *         bounds = (Var2,Var3,Var4,Var5)
- *         rowindex_lower: Var2, rowindex_upper: Var3 
- *         colindex_lower: Var4, colindex_upper: Var5
- *     leftIndex:mVar1:mVar2:Var3:Var4:Var5:Var6:mVar7
- *         triggered by "mVar1[Var3:Var4, Var5:Var6] = mVar2"
- *         the result is stored in mVar7
- *  
+ * This class implements the matrix indexing functionality inside CP.
  */
 public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 	private final LixCacheType _type;
@@ -101,12 +93,13 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 		IndexRange ixrange = new IndexRange(rl, ru, cl, cu);
 		
 		//right indexing
-		if( opcode.equalsIgnoreCase("rangeReIndex") )
+		if( opcode.equalsIgnoreCase(RightIndex.OPCODE) )
 		{
 			//update and check output dimensions
 			MatrixCharacteristics mcIn = sec.getMatrixCharacteristics(input1.getName());
 			MatrixCharacteristics mcOut = sec.getMatrixCharacteristics(output.getName());
 			mcOut.set(ru-rl+1, cu-cl+1, mcIn.getRowsPerBlock(), mcIn.getColsPerBlock());
+			mcOut.setNonZerosBound(Math.min(mcOut.getLength(), mcIn.getNonZerosBound()));
 			checkValidOutputDimensions(mcOut);
 			
 			//execute right indexing operation (partitioning-preserving if possible)
@@ -127,7 +120,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 			}
 		}
 		//left indexing
-		else if ( opcode.equalsIgnoreCase("leftIndex") || opcode.equalsIgnoreCase("mapLeftIndex"))
+		else if ( opcode.equalsIgnoreCase(LeftIndex.OPCODE) || opcode.equalsIgnoreCase("mapLeftIndex"))
 		{
 			String rddVar = (_type==LixCacheType.LEFT) ? input2.getName() : input1.getName();
 			String bcVar = (_type==LixCacheType.LEFT) ? input1.getName() : input2.getName();
@@ -198,7 +191,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 	private static MatrixBlock multiBlockIndexing(JavaPairRDD<MatrixIndexes,MatrixBlock> in1, 
 			 MatrixCharacteristics mcIn, MatrixCharacteristics mcOut, IndexRange ixrange) throws DMLRuntimeException {
 		//create list of all required matrix indexes
-		List<MatrixIndexes> filter = new ArrayList<MatrixIndexes>();
+		List<MatrixIndexes> filter = new ArrayList<>();
 		long rlix = UtilFunctions.computeBlockIndex(ixrange.rowStart, mcIn.getRowsPerBlock());
 		long ruix = UtilFunctions.computeBlockIndex(ixrange.rowEnd, mcIn.getRowsPerBlock());
 		long clix = UtilFunctions.computeBlockIndex(ixrange.colStart, mcIn.getColsPerBlock());
@@ -231,11 +224,12 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 		
 		MatrixBlock tmp = list.get(0);
 		MatrixBlock mbout = (tmp.getNumRows()==mcOut.getRows() && tmp.getNumColumns()==mcOut.getCols()) ? 
-				tmp : tmp.sliceOperations( //reference full block or slice out sub-block
+				tmp : tmp.slice( //reference full block or slice out sub-block
 				UtilFunctions.computeCellInBlock(ixrange.rowStart, mcIn.getRowsPerBlock()), 
 				UtilFunctions.computeCellInBlock(ixrange.rowEnd, mcIn.getRowsPerBlock()), 
 				UtilFunctions.computeCellInBlock(ixrange.colStart, mcIn.getColsPerBlock()), 
 				UtilFunctions.computeCellInBlock(ixrange.colEnd, mcIn.getColsPerBlock()), new MatrixBlock());
+		mbout.examSparsity();
 		return mbout;
 	}
 	
@@ -332,8 +326,8 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 		public Iterator<Tuple2<MatrixIndexes, MatrixBlock>> call(Tuple2<MatrixIndexes, MatrixBlock> rightKV) 
 			throws Exception 
 		{
-			IndexedMatrixValue in = SparkUtils.toIndexedMatrixBlock(rightKV);			
-			ArrayList<IndexedMatrixValue> out = new ArrayList<IndexedMatrixValue>();
+			IndexedMatrixValue in = SparkUtils.toIndexedMatrixBlock(rightKV);
+			ArrayList<IndexedMatrixValue> out = new ArrayList<>();
 			OperationsOnMatrixValues.performShift(in, _ixrange, _brlen, _bclen, _rlen, _clen, out);
 			return SparkUtils.fromIndexedMatrixBlock(out).iterator();
 		}		
@@ -369,7 +363,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 			}
 			
 			MatrixBlock zeroBlk = (MatrixBlock) kv._2.zeroOutOperations(new MatrixBlock(), range, _complement);
-			return new Tuple2<MatrixIndexes, MatrixBlock>(kv._1, zeroBlk);
+			return new Tuple2<>(kv._1, zeroBlk);
 		}
 	}
 
@@ -429,7 +423,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 					MatrixBlock tmp = left.leftIndexingOperations(right, 
 							rl, ru, cl, cu, new MatrixBlock(), UpdateType.COPY);
 					
-					return new Tuple2<MatrixIndexes, MatrixBlock>(ix, tmp);
+					return new Tuple2<>(ix, tmp);
 				}
 				else //LixCacheType.RIGHT
 				{
@@ -446,7 +440,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 					long rhs_cu = rhs_cl + (lhs_cu - lhs_cl);
 					
 					// Provide global zero-based index to sliceOperations
-					MatrixBlock slicedRHSMatBlock = _binput.sliceOperations(rhs_rl, rhs_ru, rhs_cl, rhs_cu, new MatrixBlock());
+					MatrixBlock slicedRHSMatBlock = _binput.slice(rhs_rl, rhs_ru, rhs_cl, rhs_cu, new MatrixBlock());
 					
 					// Provide local zero-based index to leftIndexingOperations
 					int lhs_lrl = UtilFunctions.computeCellInBlock(lhs_rl, _brlen);
@@ -454,7 +448,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 					int lhs_lcl = UtilFunctions.computeCellInBlock(lhs_cl, _bclen);
 					int lhs_lcu = UtilFunctions.computeCellInBlock(lhs_cu, _bclen);
 					MatrixBlock ret = arg._2.leftIndexingOperations(slicedRHSMatBlock, lhs_lrl, lhs_lru, lhs_lcl, lhs_lcu, new MatrixBlock(), UpdateType.COPY);
-					return new Tuple2<MatrixIndexes, MatrixBlock>(arg._1, ret);
+					return new Tuple2<>(arg._1, ret);
 				}
 			}
 		}
@@ -498,11 +492,11 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 			//create return matrix block (via shallow copy or slice)
 			if( lrl == 0 && lru == block.getNumRows()-1
 				&& lcl == 0 && lcu == block.getNumColumns()-1 ) {
-				return new Tuple2<MatrixIndexes, MatrixBlock>(ixOut, block);
+				return new Tuple2<>(ixOut, block);
 			}
 			else {
-				return new Tuple2<MatrixIndexes, MatrixBlock>(ixOut, 
-					block.sliceOperations(lrl, lru, lcl, lcu, new MatrixBlock()));
+				return new Tuple2<>(ixOut, 
+					block.slice(lrl, lru, lcl, lcu, new MatrixBlock()));
 			}
 		}		
 	}
@@ -526,7 +520,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 			throws Exception 
 		{	
 			IndexedMatrixValue in = SparkUtils.toIndexedMatrixBlock(kv);
-			ArrayList<IndexedMatrixValue> outlist = new ArrayList<IndexedMatrixValue>();
+			ArrayList<IndexedMatrixValue> outlist = new ArrayList<>();
 			OperationsOnMatrixValues.performSlice(in, _ixrange, _brlen, _bclen, outlist);
 			return SparkUtils.fromIndexedMatrixBlock(outlist).iterator();
 		}		
@@ -554,7 +548,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 			throws Exception 
 		{	
 			IndexedMatrixValue in = new IndexedMatrixValue(kv._1(), kv._2());
-			ArrayList<IndexedMatrixValue> outlist = new ArrayList<IndexedMatrixValue>();
+			ArrayList<IndexedMatrixValue> outlist = new ArrayList<>();
 			OperationsOnMatrixValues.performSlice(in, _ixrange, _brlen, _bclen, outlist);
 			return SparkUtils.fromIndexedMatrixBlock(outlist.get(0));
 		}		
@@ -593,12 +587,12 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 			{
 				IndexedMatrixValue in = SparkUtils.toIndexedMatrixBlock(arg);
 				
-				ArrayList<IndexedMatrixValue> outlist = new ArrayList<IndexedMatrixValue>();
+				ArrayList<IndexedMatrixValue> outlist = new ArrayList<>();
 				OperationsOnMatrixValues.performSlice(in, _ixrange, _brlen, _bclen, outlist);
 				
 				assert(outlist.size() == 1); //1-1 row/column block indexing
 				return SparkUtils.fromIndexedMatrixBlock(outlist.get(0));
-			}			
+			}
 		}
 	}
 	
@@ -615,7 +609,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 			JavaPairRDD<MatrixIndexes,MatrixBlock> in, List<MatrixIndexes> filter )
 	{
 		//build hashset of required partition ids
-		HashSet<Integer> flags = new HashSet<Integer>();
+		HashSet<Integer> flags = new HashSet<>();
 		Partitioner partitioner = in.rdd().partitioner().get();
 		for( MatrixIndexes key : filter )
 			flags.add(partitioner.getPartition(key));
@@ -626,7 +620,7 @@ public class MatrixIndexingSPInstruction extends IndexingSPInstruction {
 				PartitionPruningRDD.create(in.rdd(), f);
 
 		//wrap output into java pair rdd
-		return new JavaPairRDD<MatrixIndexes,MatrixBlock>(ppRDD, 
+		return new JavaPairRDD<>(ppRDD, 
 				ClassManifestFactory.fromClass(MatrixIndexes.class), 
 				ClassManifestFactory.fromClass(MatrixBlock.class));
 	}
