@@ -30,6 +30,7 @@ import java.lang.ref.SoftReference;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.HashMap;
+import java.util.HashSet;
 import java.util.Iterator;
 import java.util.Map;
 
@@ -233,7 +234,7 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 	 * @return map of column name keys and id values
 	 */
 	public Map<String,Integer> getColumnNameIDMap() {
-		Map<String, Integer> ret = new HashMap<String, Integer>();
+		Map<String, Integer> ret = new HashMap<>();
 		for( int j=0; j<getNumColumns(); j++ )
 			ret.put(getColumnName(j), j+1);
 		return ret;	
@@ -471,9 +472,9 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		for( int j=0; j<ncol; j++ )
 			tmpData[j] = new DoubleArray(cols[j]);
 		_colnames = empty ? null : (String[]) ArrayUtils.addAll(getColumnNames(), 
-				createColNames(getNumColumns(), ncol)); //before schema modification
+			createColNames(getNumColumns(), ncol)); //before schema modification
 		_schema = empty ? tmpSchema : (ValueType[]) ArrayUtils.addAll(_schema, tmpSchema); 
-		_coldata = empty ? tmpData : (Array[]) ArrayUtils.addAll(_coldata, tmpData);		
+		_coldata = empty ? tmpData : (Array[]) ArrayUtils.addAll(_coldata, tmpData);
 		_numRows = cols[0].length;
 	}
 
@@ -553,6 +554,20 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		return new ObjectRowIterator(0, _numRows);
 	}
 	
+	/**
+	 * Get a row iterator over the frame where all fields are encoded
+	 * as boxed objects according to the value types of the provided
+	 * target schema.
+	 * 
+	 * @param schema target schema of objects
+	 * @return object array iterator
+	 */
+	public Iterator<Object[]> getObjectRowIterator(ValueType[] schema) {
+		ObjectRowIterator iter = new ObjectRowIterator(0, _numRows);
+		iter.setSchema(schema);
+		return iter;
+	}
+	
 	/**
 	 * Get a row iterator over the frame where all selected fields are 
 	 * encoded as boxed objects according to their value types.  
@@ -739,15 +754,24 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 	
 	@Override
 	public boolean isShallowSerialize() {
+		return isShallowSerialize(false);
+	}
+	
+	@Override
+	public boolean isShallowSerialize(boolean inclConvert) {
 		//shallow serialize if non-string schema because a frame block
 		//is always dense but strings have large array overhead per cell
 		boolean ret = true;
 		for( int j=0; j<_schema.length && ret; j++ )
 			ret &= (_schema[j] != ValueType.STRING);
-		
 		return ret;
 	}
 	
+	@Override 
+	public void toShallowSerializeBlock() {
+		//do nothing (not applicable).
+	}
+	
 	@Override
 	public void compactEmptyBlock() {
 		//do nothing
@@ -759,7 +783,7 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 	 * @param value string value
 	 * @return in-memory size of string value
 	 */
-	private long getInMemoryStringSize(String value) {
+	private static long getInMemoryStringSize(String value) {
 		if( value == null )
 			return 0;
 		return 16 + 4 + 8 //object, hash, array ref
@@ -824,10 +848,10 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		return ret;
 	}
 
-	public FrameBlock sliceOperations(IndexRange ixrange, FrameBlock ret) 
+	public FrameBlock slice(IndexRange ixrange, FrameBlock ret) 
 		throws DMLRuntimeException
 	{
-		return sliceOperations(
+		return slice(
 				(int)ixrange.rowStart, (int)ixrange.rowEnd,
 				(int)ixrange.colStart, (int)ixrange.colEnd, ret);
 	}
@@ -844,7 +868,7 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 	 * @return frame block
 	 * @throws DMLRuntimeException if DMLRuntimeException occurs
 	 */
-	public FrameBlock sliceOperations(int rl, int ru, int cl, int cu, CacheBlock retCache) 
+	public FrameBlock slice(int rl, int ru, int cl, int cu, CacheBlock retCache) 
 		throws DMLRuntimeException
 	{
 		FrameBlock ret = (FrameBlock)retCache;
@@ -899,7 +923,7 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 	}
 	
 	
-	public void sliceOperations(ArrayList<Pair<Long,FrameBlock>> outlist, IndexRange range, int rowCut)
+	public void slice(ArrayList<Pair<Long,FrameBlock>> outlist, IndexRange range, int rowCut)
 	{
 		FrameBlock top=null, bottom=null;
 		Iterator<Pair<Long,FrameBlock>> p=outlist.iterator();
@@ -944,7 +968,7 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 	 * @return frame block
 	 * @throws DMLRuntimeException if DMLRuntimeException occurs
 	 */
-	public FrameBlock appendOperations( FrameBlock that, FrameBlock ret, boolean cbind )
+	public FrameBlock append( FrameBlock that, FrameBlock ret, boolean cbind )
 		throws DMLRuntimeException
 	{
 		if( cbind ) //COLUMN APPEND
@@ -965,10 +989,12 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 			ret._colnames = (String[]) ArrayUtils.addAll(getColumnNames(), that.getColumnNames());
 			ret._colmeta = (ColumnMetadata[]) ArrayUtils.addAll(_colmeta, that._colmeta);
 			
-			//concatenate column data (w/ deep copy to prevent side effects)
+			//check and enforce unique columns names
+			if( !Arrays.stream(ret._colnames).allMatch(new HashSet<>()::add) )
+				ret._colnames = createColNames(ret.getNumColumns());
+			
+			//concatenate column data (w/ shallow copy which is safe due to copy on write semantics)
 			ret._coldata = (Array[]) ArrayUtils.addAll(_coldata, that._coldata);
-			for( int i=0; i<ret.getNumColumns(); i++ )
-				ret._coldata[i] = ret._coldata[i].clone();
 		}
 		else //ROW APPEND
 		{
@@ -992,7 +1018,7 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 			ret._coldata = new Array[getNumColumns()];
 			for( int j=0; j<getNumColumns(); j++ )
 				ret._coldata[j] = _coldata[j].clone();
-			Iterator<Object[]> iter = that.getObjectRowIterator();
+			Iterator<Object[]> iter = that.getObjectRowIterator(_schema);
 			while( iter.hasNext() )
 				ret.appendRow(iter.next());
 		}
@@ -1044,7 +1070,7 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		}
 		
 		//construct recode map
-		HashMap<String,Long> map = new HashMap<String,Long>();
+		HashMap<String,Long> map = new HashMap<>();
 		Array ldata = _coldata[col]; 
 		for( int i=0; i<getNumRows(); i++ ) {
 			Object val = ldata.get(i);
@@ -1221,6 +1247,8 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 	}
 
 	private class ObjectRowIterator extends RowIterator<Object> {
+		private ValueType[] _tgtSchema = null;
+		
 		public ObjectRowIterator(int rl, int ru) {
 			super(rl, ru);
 		}
@@ -1229,6 +1257,10 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 			super(rl, ru, cols);
 		}
 		
+		public void setSchema(ValueType[] schema) {
+			_tgtSchema = schema;
+		}
+		
 		@Override
 		protected Object[] createRow(int size) {
 			return new Object[size];
@@ -1237,10 +1269,17 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		@Override
 		public Object[] next( ) {
 			for( int j=0; j<_cols.length; j++ )
-				_curRow[j] = get(_curPos, _cols[j]-1);
+				_curRow[j] = getValue(_curPos, _cols[j]-1);
 			_curPos++;
 			return _curRow;
 		}
+		
+		private Object getValue(int i, int j) {
+			Object val = get(i, j);
+			if( _tgtSchema != null )
+				val = UtilFunctions.objectToObject(_tgtSchema[j], val);
+			return val;
+		}
 	}
 	
 	///////
@@ -1265,6 +1304,7 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		public abstract void setNz(int rl, int ru, Array value);
 		public abstract void append(String value);
 		public abstract void append(T value);
+		@Override
 		public abstract Array clone();
 		public abstract Array slice(int rl, int ru);
 		public abstract void reset(int size); 
@@ -1276,25 +1316,31 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		public StringArray(String[] data) {
 			_data = data;
 			_size = _data.length;
-		}		
+		}
+		@Override
 		public String get(int index) {
 			return _data[index];
 		}
+		@Override
 		public void set(int index, String value) {
 			_data[index] = value;
 		}
+		@Override
 		public void set(int rl, int ru, Array value) {
 			set(rl, ru, value, 0);
 		}
+		@Override
 		public void set(int rl, int ru, Array value, int rlSrc) {
 			System.arraycopy(((StringArray)value)._data, rlSrc, _data, rl, ru-rl+1);
 		}
+		@Override
 		public void setNz(int rl, int ru, Array value) {
 			String[] data2 = ((StringArray)value)._data;
 			for( int i=rl; i<ru+1; i++ )
 				if( data2[i]!=null )
 					_data[i] = data2[i];
 		}
+		@Override
 		public void append(String value) {
 			if( _data.length <= _size )
 				_data = Arrays.copyOf(_data, newSize());
@@ -1311,12 +1357,15 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 				_data[i] = (!tmp.isEmpty()) ? tmp : null;
 			}
 		}
+		@Override
 		public Array clone() {
 			return new StringArray(Arrays.copyOf(_data, _size));
 		}
+		@Override
 		public Array slice(int rl, int ru) {
 			return new StringArray(Arrays.copyOfRange(_data,rl,ru+1));
 		}
+		@Override
 		public void reset(int size) {
 			if( _data.length < size )
 				_data = new String[size];
@@ -1330,28 +1379,35 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		public BooleanArray(boolean[] data) {
 			_data = data;
 			_size = _data.length;
-		}		
+		}
+		@Override
 		public Boolean get(int index) {
 			return _data[index];
 		}
+		@Override
 		public void set(int index, Boolean value) {
 			_data[index] = (value!=null) ? value : false;
 		}
+		@Override
 		public void set(int rl, int ru, Array value) {
 			set(rl, ru, value, 0);
 		}
+		@Override
 		public void set(int rl, int ru, Array value, int rlSrc) {
 			System.arraycopy(((BooleanArray)value)._data, rlSrc, _data, rl, ru-rl+1);
 		}
+		@Override
 		public void setNz(int rl, int ru, Array value) {
 			boolean[] data2 = ((BooleanArray)value)._data;
 			for( int i=rl; i<ru+1; i++ )
 				if( data2[i] )
 					_data[i] = data2[i];
 		}
+		@Override
 		public void append(String value) {
 			append(Boolean.parseBoolean(value));
 		}
+		@Override
 		public void append(Boolean value) {
 			if( _data.length <= _size )
 				_data = Arrays.copyOf(_data, newSize());
@@ -1366,12 +1422,15 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 			for( int i=0; i<_size; i++ )
 				_data[i] = in.readBoolean();
 		}
+		@Override
 		public Array clone() {
 			return new BooleanArray(Arrays.copyOf(_data, _size));
 		}
+		@Override
 		public Array slice(int rl, int ru) {
 			return new BooleanArray(Arrays.copyOfRange(_data,rl,ru+1));
 		}
+		@Override
 		public void reset(int size) {
 			if( _data.length < size )
 				_data = new boolean[size];
@@ -1385,28 +1444,35 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		public LongArray(long[] data) {
 			_data = data;
 			_size = _data.length;
-		}		
+		}
+		@Override
 		public Long get(int index) {
 			return _data[index];
 		}
+		@Override
 		public void set(int index, Long value) {
 			_data[index] = (value!=null) ? value : 0L;
 		}
+		@Override
 		public void set(int rl, int ru, Array value) {
 			set(rl, ru, value, 0);
 		}
+		@Override
 		public void set(int rl, int ru, Array value, int rlSrc) {
 			System.arraycopy(((LongArray)value)._data, rlSrc, _data, rl, ru-rl+1);
 		}
+		@Override
 		public void setNz(int rl, int ru, Array value) {
 			long[] data2 = ((LongArray)value)._data;
 			for( int i=rl; i<ru+1; i++ )
 				if( data2[i]!=0 )
 					_data[i] = data2[i];
 		}
+		@Override
 		public void append(String value) {
 			append((value!=null)?Long.parseLong(value):null);
 		}
+		@Override
 		public void append(Long value) {
 			if( _data.length <= _size )
 				_data = Arrays.copyOf(_data, newSize());
@@ -1421,12 +1487,15 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 			for( int i=0; i<_size; i++ )
 				_data[i] = in.readLong();
 		}
+		@Override
 		public Array clone() {
 			return new LongArray(Arrays.copyOf(_data, _size));
 		}
+		@Override
 		public Array slice(int rl, int ru) {
 			return new LongArray(Arrays.copyOfRange(_data,rl,ru+1));
 		}
+		@Override
 		public void reset(int size) {
 			if( _data.length < size )
 				_data = new long[size];
@@ -1440,28 +1509,35 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 		public DoubleArray(double[] data) {
 			_data = data;
 			_size = _data.length;
-		}		
+		}
+		@Override
 		public Double get(int index) {
 			return _data[index];
 		}
+		@Override
 		public void set(int index, Double value) {
 			_data[index] = (value!=null) ? value : 0d;
 		}
+		@Override
 		public void set(int rl, int ru, Array value) {
 			set(rl,ru, value, 0);
 		}
+		@Override
 		public void set(int rl, int ru, Array value, int rlSrc) {
 			System.arraycopy(((DoubleArray)value)._data, rlSrc, _data, rl, ru-rl+1);
 		}
+		@Override
 		public void setNz(int rl, int ru, Array value) {
 			double[] data2 = ((DoubleArray)value)._data;
 			for( int i=rl; i<ru+1; i++ )
 				if( data2[i]!=0 )
 					_data[i] = data2[i];
 		}
+		@Override
 		public void append(String value) {
 			append((value!=null)?Double.parseDouble(value):null);
 		}
+		@Override
 		public void append(Double value) {
 			if( _data.length <= _size )
 				_data = Arrays.copyOf(_data, newSize());
@@ -1476,12 +1552,15 @@ public class FrameBlock implements Writable, CacheBlock, Externalizable
 			for( int i=0; i<_size; i++ )
 				_data[i] = in.readDouble();
 		}
+		@Override
 		public Array clone() {
 			return new DoubleArray(Arrays.copyOf(_data, _size));
 		}
+		@Override
 		public Array slice(int rl, int ru) {
 			return new DoubleArray(Arrays.copyOfRange(_data,rl,ru+1));
 		}
+		@Override
 		public void reset(int size) {
 			if( _data.length < size )
 				_data = new double[size];
