@@ -25,6 +25,7 @@ import java.io.IOException;
 import java.util.Arrays;
 
 import org.apache.sysml.runtime.util.SortUtils;
+import org.apache.sysml.runtime.util.UtilFunctions;
 
 /**
  * SparseBlock implementation that realizes a traditional 'compressed sparse row'
@@ -68,7 +69,7 @@ public class SparseBlockCSR extends SparseBlock
 		_values = values;
 		_size = nnz;
 	}
-	
+
 	/**
 	 * Copy constructor sparse block abstraction. 
 	 * 
@@ -108,7 +109,7 @@ public class SparseBlockCSR extends SparseBlock
 					pos += alen;
 				}
 				_ptr[i+1]=pos;
-			}			
+			}
 		}
 	}
 	
@@ -165,6 +166,33 @@ public class SparseBlockCSR extends SparseBlock
 		Arrays.fill(_ptr, rlast+1, numRows()+1, nnz);
 	}
 	
+	/**
+	 * Copy constructor for given array of column indexes, which
+	 * identifies rows by position and implies values of 1.
+	 * 
+	 * @param rows number of rows
+	 * @param nnz number of non-zeros
+	 * @param colInd column indexes
+	 */
+	public SparseBlockCSR(int rows, int nnz, int[] colInd) {
+		_ptr = new int[rows+1];
+		_indexes = (rows==nnz) ? colInd : new int[nnz];
+		_values = new double[nnz];
+		Arrays.fill(_values, 1);
+		_size = nnz;
+		
+		//single-pass construction of row pointers
+		//and copy of column indexes if necessary
+		for(int i=0, pos=0; i<rows; i++) {
+			if( colInd[i] >= 0 ) {
+				if( rows > nnz )
+					_indexes[pos] = colInd[i];
+				pos++;
+			}
+			_ptr[i+1] = pos;
+		}
+	}
+	
 	/**
 	 * Initializes the CSR sparse block from an ordered input
 	 * stream of ultra-sparse ijv triples. 
@@ -249,6 +277,34 @@ public class SparseBlockCSR extends SparseBlock
 		return (long) Math.min(size, Long.MAX_VALUE);
 	}
 	
+
+	/**
+	 * Get raw access to underlying array of row pointers
+	 * For use in GPU code
+	 * @return array of row pointers
+	 */
+	public int[] rowPointers() {
+		return _ptr;
+	}
+	
+	/** 
+	 * Get raw access to underlying array of column indices
+	 * For use in GPU code
+	 * @return array of column indexes
+	 */
+	public int[] indexes() {
+		return indexes(0);
+	}
+	
+	/**
+	 * Get raw access to underlying array of values
+	 * For use in GPU code
+	 * @return array of values
+	 */
+	public double[] values() {
+		return values(0);
+	}
+	
 	///////////////////
 	//SparseBlock implementation
 
@@ -282,6 +338,11 @@ public class SparseBlockCSR extends SparseBlock
 		return true;
 	}
 	
+	@Override
+	public boolean isAllocated(int r) {
+		return true;
+	}
+
 	@Override 
 	public void reset() {
 		if( _size > 0 ) {
@@ -332,8 +393,8 @@ public class SparseBlockCSR extends SparseBlock
 		long nnz = 0;
 		for(int i=rl; i<ru; i++)
 			if( !isEmpty(i) ) {
-				int start = posFIndexGTE(i, cl);
-				int end = posFIndexGTE(i, cu);
+				int start = internPosFIndexGTE(i, cl);
+				int end = internPosFIndexGTE(i, cu);
 				nnz += (start!=-1) ? (end-start) : 0;
 			}
 		return nnz;
@@ -396,14 +457,14 @@ public class SparseBlockCSR extends SparseBlock
 	@Override
 	public void set(int r, SparseRow row, boolean deep) {
 		int pos = pos(r);
-		int len = size(r);		
+		int len = size(r);
 		int alen = row.size();
 		int[] aix = row.indexes();
 		double[] avals = row.values();
 		
 		//delete existing values if necessary
 		if( len > 0 ) //incl size update
-			deleteIndexRange(r, aix[0], aix[alen-1]+1);
+			deleteIndexRange(r, aix[pos], aix[pos+len-1]+1);
 		
 		//prepare free space (allocate and shift)
 		int lsize = _size+alen;
@@ -442,20 +503,18 @@ public class SparseBlockCSR extends SparseBlock
 
 	@Override
 	public void setIndexRange(int r, int cl, int cu, double[] v, int vix, int vlen) {
-		//delete existing values in range if necessary 
+		//delete existing values in range if necessary
 		if( !isEmpty(r) )
 			deleteIndexRange(r, cl, cu);
 		
 		//determine input nnz
-		int lnnz = 0;
-		for( int i=vix; i<vix+vlen; i++ )
-			lnnz += ( v[i] != 0 ) ? 1 : 0;
-
+		int lnnz = UtilFunctions.computeNnz(v, vix, vlen);
+		
 		//prepare free space (allocate and shift)
 		int lsize = _size+lnnz;
 		if( _values.length < lsize )
 			resize(lsize);
-		int index = posFIndexGT(r, cl);
+		int index = internPosFIndexGT(r, cl);
 		int index2 = (index>0)?index:pos(r+1);
 		shiftRightByN(index2, lnnz);
 		
@@ -469,6 +528,29 @@ public class SparseBlockCSR extends SparseBlock
 		incrPtr(r+1, lnnz);
 	}
 	
+	@Override
+	public void setIndexRange(int r, int cl, int cu, double[] v, int[] vix, int vpos, int vlen) {
+		//delete existing values in range if necessary
+		if( !isEmpty(r) )
+			deleteIndexRange(r, cl, cu);
+		
+		//prepare free space (allocate and shift)
+		int lsize = _size+vlen;
+		if( _values.length < lsize )
+			resize(lsize);
+		int index = internPosFIndexGT(r, cl);
+		int index2 = (index>0)?index:pos(r+1);
+		shiftRightByN(index2, vlen);
+		
+		//insert values
+		for( int i=vpos; i<vpos+vlen; i++ ) {
+			_indexes[ index2 ] = cl+vix[i];
+			_values[ index2 ] = v[i];
+			index2++;
+		}
+		incrPtr(r+1, vlen);
+	}
+	
 	/**
 	 * Inserts a sorted row-major array of non-zero values into the row and column 
 	 * range [rl,ru) and [cl,cu). Note: that this is a CSR-specific method to address 
@@ -486,8 +568,7 @@ public class SparseBlockCSR extends SparseBlock
 		//step 1: determine output nnz
 		int nnz = _size - (int)size(rl, ru, cl, cu);
 		if( v != null )
-			for( int i=vix; i<vix+vlen; i++ )
-				nnz += (v[i]!=0) ? 1: 0;
+			nnz += UtilFunctions.computeNnz(v, vix, vlen);
 		
 		//step 2: reallocate if necessary
 		if( _values.length < nnz )
@@ -618,12 +699,12 @@ public class SparseBlockCSR extends SparseBlock
 
 	@Override
 	public void deleteIndexRange(int r, int cl, int cu) {
-		int start = posFIndexGTE(r,cl);
+		int start = internPosFIndexGTE(r,cl);
 		if( start < 0 ) //nothing to delete 
-			return;		
+			return;
 
 		int len = size(r);
-		int end = posFIndexGTE(r, cu);
+		int end = internPosFIndexGTE(r, cu);
 		if( end < 0 ) //delete all remaining
 			end = start+len;
 		
@@ -676,6 +757,11 @@ public class SparseBlockCSR extends SparseBlock
 	
 	@Override
 	public int posFIndexLTE(int r, int c) {
+		int index = internPosFIndexLTE(r, c);
+		return (index>=0) ? index-pos(r) : index;
+	}
+	
+	private int internPosFIndexLTE(int r, int c) {
 		int pos = pos(r);
 		int len = size(r);
 		
@@ -691,6 +777,11 @@ public class SparseBlockCSR extends SparseBlock
 
 	@Override
 	public int posFIndexGTE(int r, int c) {
+		int index = internPosFIndexGTE(r, c);
+		return (index>=0) ? index-pos(r) : index;
+	}
+	
+	private int internPosFIndexGTE(int r, int c) {
 		int pos = pos(r);
 		int len = size(r);
 		
@@ -706,6 +797,11 @@ public class SparseBlockCSR extends SparseBlock
 
 	@Override
 	public int posFIndexGT(int r, int c) {
+		int index = internPosFIndexGT(r, c);
+		return (index>=0) ? index-pos(r) : index;
+	}
+	
+	private int internPosFIndexGT(int r, int c) {
 		int pos = pos(r);
 		int len = size(r);
 		
@@ -741,22 +837,72 @@ public class SparseBlockCSR extends SparseBlock
 				sb.append("\t");
 			}
 			sb.append("\n");
-		}		
+		}
 		
 		return sb.toString();
 	}
-	
+
+	@Override
+	public boolean checkValidity(int rlen, int clen, long nnz, boolean strict) {
+		//1. correct meta data
+		if( rlen < 0 || clen < 0 ) {
+			throw new RuntimeException("Invalid block dimensions: "+rlen+" "+clen);
+		}
+
+		//2. correct array lengths
+		if(_size != nnz && _ptr.length < rlen+1 && _values.length < nnz && _indexes.length < nnz ) {
+			throw new RuntimeException("Incorrect array lengths.");
+		}
+
+		//3. non-decreasing row pointers
+		for( int i=1; i<rlen; i++ ) {
+			if(_ptr[i-1] > _ptr[i] && strict)
+				throw new RuntimeException("Row pointers are decreasing at row: "+i
+					+ ", with pointers "+_ptr[i-1]+" > "+_ptr[i]);
+		}
+
+		//4. sorted column indexes per row
+		for( int i=0; i<rlen; i++ ) {
+			int apos = pos(i);
+			int alen = size(i);
+			for( int k=apos+1; k<apos+alen; k++)
+				if( _indexes[k-1] >= _indexes[k] )
+					throw new RuntimeException("Wrong sparse row ordering: "
+						+ k + " "+_indexes[k-1]+" "+_indexes[k]);
+			for( int k=apos; k<apos+alen; k++ )
+				if( _values[k] == 0 )
+					throw new RuntimeException("Wrong sparse row: zero at "
+						+ k + " at col index " + _indexes[k]);
+		}
+
+		//5. non-existing zero values
+		for( int i=0; i<_size; i++ ) {
+			if( _values[i] == 0 ) {
+				throw new RuntimeException("The values array should not contain zeros."
+					+ " The " + i + "th value is "+_values[i]);
+			}
+		}
+
+		//6. a capacity that is no larger than nnz times resize factor.
+		int capacity = _values.length;
+		if(capacity > nnz*RESIZE_FACTOR1 ) {
+			throw new RuntimeException("Capacity is larger than the nnz times a resize factor."
+				+ " Current size: "+capacity+ ", while Expected size:"+nnz*RESIZE_FACTOR1);
+		}
+
+		return true;
+	}
+
 	///////////////////////////
 	// private helper methods
 	
 	private int newCapacity(int minsize) {
 		//compute new size until minsize reached
-		double tmpCap = _values.length;
+		double tmpCap = Math.max(_values.length, 1);
 		while( tmpCap < minsize ) {
 			tmpCap *= (tmpCap <= 1024) ? 
-					RESIZE_FACTOR1 : RESIZE_FACTOR2;
+				RESIZE_FACTOR1 : RESIZE_FACTOR2;
 		}
-			
 		return (int)Math.min(tmpCap, Integer.MAX_VALUE);
 	}
 
@@ -798,7 +944,7 @@ public class SparseBlockCSR extends SparseBlock
 		insert(ix, c, v);
 	}
 
-	private void shiftRightAndInsert(int ix, int c, double v)  {		
+	private void shiftRightAndInsert(int ix, int c, double v)  {
 		//overlapping array copy (shift rhs values right by 1)
 		System.arraycopy(_indexes, ix, _indexes, ix+1, _size-ix);
 		System.arraycopy(_values, ix, _values, ix+1, _size-ix);
@@ -856,31 +1002,4 @@ public class SparseBlockCSR extends SparseBlock
 		for( int i=rl; i<rlen+1; i++ )
 			_ptr[i]-=cnt;
 	}
-	
-	/**
-	 * Get raw access to underlying array of row pointers
-	 * For use in GPU code
-	 * @return array of row pointers
-	 */
-	public int[] rowPointers() {
-		return _ptr;
-	}
-	
-	/** 
-	 * Get raw access to underlying array of column indices
-	 * For use in GPU code
-	 * @return array of column indexes
-	 */
-	public int[] indexes() {
-		return _indexes;
-	}
-	
-	/**
-	 * Get raw access to underlying array of values
-	 * For use in GPU code
-	 * @return array of values
-	 */
-	public double[] values() {
-		return _values;
-	}
 }
