@@ -35,7 +35,6 @@ import java.io.IOException;
 import java.io.InputStreamReader;
 import java.io.OutputStreamWriter;
 import java.io.PrintWriter;
-import java.text.DecimalFormat;
 import java.text.NumberFormat;
 import java.util.ArrayList;
 import java.util.Collections;
@@ -571,12 +570,8 @@ public class TestUtils
 			}
 			else if ( count > 1 ) {
 				File tmp = new File(csvFile+"_temp.csv");
-				OutputStreamWriter out = null;
-
-				try {
-					out = new OutputStreamWriter(new FileOutputStream(tmp),
-							"UTF-8");
 
+				try( OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(tmp), "UTF-8") ) {
 					// Directory listing may contain .crc files or may be in the
 					// wrong order. Sanitize the list of names.
 					ArrayList<String> partNames = new ArrayList<String>();
@@ -595,8 +590,6 @@ public class TestUtils
 								"UTF-8");
 						out.append(fileContents);
 					}
-				} finally {
-					IOUtilFunctions.closeSilently(out);
 				}
 				
 				csvFile = tmp.getCanonicalPath();
@@ -2011,7 +2004,7 @@ public class TestUtils
 	 * @return string representation
 	 */
 	public static String getStringRepresentationForDouble(double value) {
-		NumberFormat nf = DecimalFormat.getInstance(new Locale("EN"));
+		NumberFormat nf = NumberFormat.getInstance(new Locale("EN"));
 		nf.setGroupingUsed(false);
 		nf.setMinimumFractionDigits(1);
 		nf.setMaximumFractionDigits(20);
@@ -2051,29 +2044,33 @@ public class TestUtils
 		return matrix;
 	}
 	
-	/**
-	 * 
-	 * @param data
-	 * @return
-	 */
-	public static double[][] round(double[][] data) 
-	{
+	public static double[][] round(double[][] data) {
 		for(int i=0; i<data.length; i++)
 			for(int j=0; j<data[i].length; j++)
 				data[i][j]=Math.round(data[i][j]);
-		
 		return data;
 	}
 	
-	/**
-	 * 
-	 * @param data
-	 * @param rows
-	 * @param cols
-	 * @return
-	 */
-	public static double sum(double[][] data, int rows, int cols)
-	{
+	public static double[][] round(double[][] data, int col) {
+		for(int i=0; i<data.length; i++)
+			data[i][col]=Math.round(data[i][col]);
+		return data;
+	}
+	
+	public static double[][] floor(double[][] data) {
+		for(int i=0; i<data.length; i++)
+			for(int j=0; j<data[i].length; j++)
+				data[i][j]=Math.floor(data[i][j]);
+		return data;
+	}
+	
+	public static double[][] floor(double[][] data, int col) {
+		for(int i=0; i<data.length; i++)
+			data[i][col]=Math.floor(data[i][col]);
+		return data;
+	}
+	
+	public static double sum(double[][] data, int rows, int cols) {
 		double sum = 0;
 		for (int i = 0; i< rows; i++){
 			for (int j = 0; j < cols; j++){
@@ -2082,4 +2079,11 @@ public class TestUtils
 		}
 		return sum;
 	}
+	
+	public static long computeNNZ(double[][] data) {
+		long nnz = 0;
+		for(int i=0; i<data.length; i++)
+			nnz += UtilFunctions.computeNnz(data[i], 0, data[i].length);
+		return nnz;
+	}
 }
